/* Autogen header for Unicorn Engine - DONOT MODIFY */
#ifndef UNICORN_AUTOGEN_x86_64_H
#define UNICORN_AUTOGEN_x86_64_H
#ifndef UNICORN_ARCH_POSTFIX
#define UNICORN_ARCH_POSTFIX _x86_64
#endif
#define gen_helper_check_exit_request gen_helper_check_exit_request_x86_64
#define unicorn_fill_tlb unicorn_fill_tlb_x86_64
#define reg_read reg_read_x86_64
#define reg_write reg_write_x86_64
#define uc_init uc_init_x86_64
#define uc_add_inline_hook uc_add_inline_hook_x86_64
#define uc_del_inline_hook uc_del_inline_hook_x86_64
#define tb_invalidate_phys_range tb_invalidate_phys_range_x86_64
#define use_idiv_instructions use_idiv_instructions_x86_64
#define arm_arch arm_arch_x86_64
#define tb_target_set_jmp_target tb_target_set_jmp_target_x86_64
#define have_bmi1 have_bmi1_x86_64
#define have_popcnt have_popcnt_x86_64
#define have_avx1 have_avx1_x86_64
#define have_avx2 have_avx2_x86_64
#define have_isa have_isa_x86_64
#define have_altivec have_altivec_x86_64
#define have_vsx have_vsx_x86_64
#define flush_icache_range flush_icache_range_x86_64
#define s390_facilities s390_facilities_x86_64
#define tcg_dump_op tcg_dump_op_x86_64
#define tcg_dump_ops tcg_dump_ops_x86_64
#define tcg_gen_and_i64 tcg_gen_and_i64_x86_64
#define tcg_gen_discard_i64 tcg_gen_discard_i64_x86_64
#define tcg_gen_ld16s_i64 tcg_gen_ld16s_i64_x86_64
#define tcg_gen_ld16u_i64 tcg_gen_ld16u_i64_x86_64
#define tcg_gen_ld32s_i64 tcg_gen_ld32s_i64_x86_64
#define tcg_gen_ld32u_i64 tcg_gen_ld32u_i64_x86_64
#define tcg_gen_ld8s_i64 tcg_gen_ld8s_i64_x86_64
#define tcg_gen_ld8u_i64 tcg_gen_ld8u_i64_x86_64
#define tcg_gen_ld_i64 tcg_gen_ld_i64_x86_64
#define tcg_gen_mov_i64 tcg_gen_mov_i64_x86_64
#define tcg_gen_movi_i64 tcg_gen_movi_i64_x86_64
#define tcg_gen_mul_i64 tcg_gen_mul_i64_x86_64
#define tcg_gen_or_i64 tcg_gen_or_i64_x86_64
#define tcg_gen_sar_i64 tcg_gen_sar_i64_x86_64
#define tcg_gen_shl_i64 tcg_gen_shl_i64_x86_64
#define tcg_gen_shr_i64 tcg_gen_shr_i64_x86_64
#define tcg_gen_st_i64 tcg_gen_st_i64_x86_64
#define tcg_gen_xor_i64 tcg_gen_xor_i64_x86_64
#define cpu_icount_to_ns cpu_icount_to_ns_x86_64
#define cpu_is_stopped cpu_is_stopped_x86_64
#define cpu_get_ticks cpu_get_ticks_x86_64
#define cpu_get_clock cpu_get_clock_x86_64
#define cpu_resume cpu_resume_x86_64
#define qemu_init_vcpu qemu_init_vcpu_x86_64
#define cpu_stop_current cpu_stop_current_x86_64
#define resume_all_vcpus resume_all_vcpus_x86_64
#define vm_start vm_start_x86_64
#define address_space_dispatch_compact address_space_dispatch_compact_x86_64
#define flatview_translate flatview_translate_x86_64
#define flatview_copy flatview_copy_x86_64
#define address_space_translate_for_iotlb address_space_translate_for_iotlb_x86_64
#define qemu_get_cpu qemu_get_cpu_x86_64
#define cpu_address_space_init cpu_address_space_init_x86_64
#define cpu_get_address_space cpu_get_address_space_x86_64
#define cpu_exec_unrealizefn cpu_exec_unrealizefn_x86_64
#define cpu_exec_initfn cpu_exec_initfn_x86_64
#define cpu_exec_realizefn cpu_exec_realizefn_x86_64
#define tb_invalidate_phys_addr tb_invalidate_phys_addr_x86_64
#define cpu_watchpoint_insert cpu_watchpoint_insert_x86_64
#define cpu_watchpoint_remove_by_ref cpu_watchpoint_remove_by_ref_x86_64
#define cpu_watchpoint_remove_all cpu_watchpoint_remove_all_x86_64
#define cpu_watchpoint_address_matches cpu_watchpoint_address_matches_x86_64
#define cpu_breakpoint_insert cpu_breakpoint_insert_x86_64
#define cpu_breakpoint_remove cpu_breakpoint_remove_x86_64
#define cpu_breakpoint_remove_by_ref cpu_breakpoint_remove_by_ref_x86_64
#define cpu_breakpoint_remove_all cpu_breakpoint_remove_all_x86_64
#define cpu_abort cpu_abort_x86_64
#define cpu_physical_memory_test_and_clear_dirty cpu_physical_memory_test_and_clear_dirty_x86_64
#define memory_region_section_get_iotlb memory_region_section_get_iotlb_x86_64
#define flatview_add_to_dispatch flatview_add_to_dispatch_x86_64
#define qemu_ram_get_host_addr qemu_ram_get_host_addr_x86_64
#define qemu_ram_get_offset qemu_ram_get_offset_x86_64
#define qemu_ram_get_used_length qemu_ram_get_used_length_x86_64
#define qemu_ram_is_shared qemu_ram_is_shared_x86_64
#define qemu_ram_pagesize qemu_ram_pagesize_x86_64
#define qemu_ram_alloc_from_ptr qemu_ram_alloc_from_ptr_x86_64
#define qemu_ram_alloc qemu_ram_alloc_x86_64
#define qemu_ram_free qemu_ram_free_x86_64
#define qemu_map_ram_ptr qemu_map_ram_ptr_x86_64
#define qemu_ram_block_host_offset qemu_ram_block_host_offset_x86_64
#define qemu_ram_block_from_host qemu_ram_block_from_host_x86_64
#define qemu_ram_addr_from_host qemu_ram_addr_from_host_x86_64
#define cpu_check_watchpoint cpu_check_watchpoint_x86_64
#define iotlb_to_section iotlb_to_section_x86_64
#define address_space_dispatch_new address_space_dispatch_new_x86_64
#define address_space_dispatch_free address_space_dispatch_free_x86_64
#define address_space_dispatch_clear address_space_dispatch_clear_x86_64
#define flatview_read_continue flatview_read_continue_x86_64
#define address_space_read_full address_space_read_full_x86_64
#define address_space_write address_space_write_x86_64
#define address_space_rw address_space_rw_x86_64
#define cpu_physical_memory_rw cpu_physical_memory_rw_x86_64
#define address_space_write_rom address_space_write_rom_x86_64
#define cpu_flush_icache_range cpu_flush_icache_range_x86_64
#define cpu_exec_init_all cpu_exec_init_all_x86_64
#define address_space_access_valid address_space_access_valid_x86_64
#define address_space_map address_space_map_x86_64
#define address_space_unmap address_space_unmap_x86_64
#define cpu_physical_memory_map cpu_physical_memory_map_x86_64
#define cpu_physical_memory_unmap cpu_physical_memory_unmap_x86_64
#define cpu_memory_rw_debug cpu_memory_rw_debug_x86_64
#define qemu_target_page_size qemu_target_page_size_x86_64
#define qemu_target_page_bits qemu_target_page_bits_x86_64
#define qemu_target_page_bits_min qemu_target_page_bits_min_x86_64
#define target_words_bigendian target_words_bigendian_x86_64
#define cpu_physical_memory_is_io cpu_physical_memory_is_io_x86_64
#define ram_block_discard_range ram_block_discard_range_x86_64
#define ramblock_is_pmem ramblock_is_pmem_x86_64
#define page_size_init page_size_init_x86_64
#define set_preferred_target_page_bits set_preferred_target_page_bits_x86_64
#define finalize_target_page_bits finalize_target_page_bits_x86_64
#define cpu_outb cpu_outb_x86_64
#define cpu_outw cpu_outw_x86_64
#define cpu_outl cpu_outl_x86_64
#define cpu_inb cpu_inb_x86_64
#define cpu_inw cpu_inw_x86_64
#define cpu_inl cpu_inl_x86_64
#define memory_map memory_map_x86_64
#define memory_map_io memory_map_io_x86_64
#define memory_map_ptr memory_map_ptr_x86_64
#define memory_cow memory_cow_x86_64
#define memory_unmap memory_unmap_x86_64
#define memory_moveout memory_moveout_x86_64
#define memory_movein memory_movein_x86_64
#define memory_free memory_free_x86_64
#define flatview_unref flatview_unref_x86_64
#define address_space_get_flatview address_space_get_flatview_x86_64
#define memory_region_transaction_begin memory_region_transaction_begin_x86_64
#define memory_region_transaction_commit memory_region_transaction_commit_x86_64
#define memory_region_init memory_region_init_x86_64
#define memory_region_access_valid memory_region_access_valid_x86_64
#define memory_region_dispatch_read memory_region_dispatch_read_x86_64
#define memory_region_dispatch_write memory_region_dispatch_write_x86_64
#define memory_region_init_io memory_region_init_io_x86_64
#define memory_region_init_ram_ptr memory_region_init_ram_ptr_x86_64
#define memory_region_size memory_region_size_x86_64
#define memory_region_set_readonly memory_region_set_readonly_x86_64
#define memory_region_get_ram_ptr memory_region_get_ram_ptr_x86_64
#define memory_region_from_host memory_region_from_host_x86_64
#define memory_region_get_ram_addr memory_region_get_ram_addr_x86_64
#define memory_region_add_subregion memory_region_add_subregion_x86_64
#define memory_region_del_subregion memory_region_del_subregion_x86_64
#define memory_region_add_subregion_overlap memory_region_add_subregion_overlap_x86_64
#define memory_region_find memory_region_find_x86_64
#define memory_region_filter_subregions memory_region_filter_subregions_x86_64
#define memory_listener_register memory_listener_register_x86_64
#define memory_listener_unregister memory_listener_unregister_x86_64
#define address_space_remove_listeners address_space_remove_listeners_x86_64
#define address_space_init address_space_init_x86_64
#define address_space_destroy address_space_destroy_x86_64
#define memory_region_init_ram memory_region_init_ram_x86_64
#define memory_mapping_list_add_merge_sorted memory_mapping_list_add_merge_sorted_x86_64
#define find_memory_mapping find_memory_mapping_x86_64
#define exec_inline_op exec_inline_op_x86_64
#define floatx80_default_nan floatx80_default_nan_x86_64
#define float_raise float_raise_x86_64
#define float16_is_quiet_nan float16_is_quiet_nan_x86_64
#define float16_is_signaling_nan float16_is_signaling_nan_x86_64
#define float32_is_quiet_nan float32_is_quiet_nan_x86_64
#define float32_is_signaling_nan float32_is_signaling_nan_x86_64
#define float64_is_quiet_nan float64_is_quiet_nan_x86_64
#define float64_is_signaling_nan float64_is_signaling_nan_x86_64
#define floatx80_is_quiet_nan floatx80_is_quiet_nan_x86_64
#define floatx80_is_signaling_nan floatx80_is_signaling_nan_x86_64
#define floatx80_silence_nan floatx80_silence_nan_x86_64
#define propagateFloatx80NaN propagateFloatx80NaN_x86_64
#define float128_is_quiet_nan float128_is_quiet_nan_x86_64
#define float128_is_signaling_nan float128_is_signaling_nan_x86_64
#define float128_silence_nan float128_silence_nan_x86_64
#define float16_add float16_add_x86_64
#define float16_sub float16_sub_x86_64
#define float32_add float32_add_x86_64
#define float32_sub float32_sub_x86_64
#define float64_add float64_add_x86_64
#define float64_sub float64_sub_x86_64
#define float16_mul float16_mul_x86_64
#define float32_mul float32_mul_x86_64
#define float64_mul float64_mul_x86_64
#define float16_muladd float16_muladd_x86_64
#define float32_muladd float32_muladd_x86_64
#define float64_muladd float64_muladd_x86_64
#define float16_div float16_div_x86_64
#define float32_div float32_div_x86_64
#define float64_div float64_div_x86_64
#define float16_to_float32 float16_to_float32_x86_64
#define float16_to_float64 float16_to_float64_x86_64
#define float32_to_float16 float32_to_float16_x86_64
#define float32_to_float64 float32_to_float64_x86_64
#define float64_to_float16 float64_to_float16_x86_64
#define float64_to_float32 float64_to_float32_x86_64
#define float16_round_to_int float16_round_to_int_x86_64
#define float32_round_to_int float32_round_to_int_x86_64
#define float64_round_to_int float64_round_to_int_x86_64
#define float16_to_int16_scalbn float16_to_int16_scalbn_x86_64
#define float16_to_int32_scalbn float16_to_int32_scalbn_x86_64
#define float16_to_int64_scalbn float16_to_int64_scalbn_x86_64
#define float32_to_int16_scalbn float32_to_int16_scalbn_x86_64
#define float32_to_int32_scalbn float32_to_int32_scalbn_x86_64
#define float32_to_int64_scalbn float32_to_int64_scalbn_x86_64
#define float64_to_int16_scalbn float64_to_int16_scalbn_x86_64
#define float64_to_int32_scalbn float64_to_int32_scalbn_x86_64
#define float64_to_int64_scalbn float64_to_int64_scalbn_x86_64
#define float16_to_int16 float16_to_int16_x86_64
#define float16_to_int32 float16_to_int32_x86_64
#define float16_to_int64 float16_to_int64_x86_64
#define float32_to_int16 float32_to_int16_x86_64
#define float32_to_int32 float32_to_int32_x86_64
#define float32_to_int64 float32_to_int64_x86_64
#define float64_to_int16 float64_to_int16_x86_64
#define float64_to_int32 float64_to_int32_x86_64
#define float64_to_int64 float64_to_int64_x86_64
#define float16_to_int16_round_to_zero float16_to_int16_round_to_zero_x86_64
#define float16_to_int32_round_to_zero float16_to_int32_round_to_zero_x86_64
#define float16_to_int64_round_to_zero float16_to_int64_round_to_zero_x86_64
#define float32_to_int16_round_to_zero float32_to_int16_round_to_zero_x86_64
#define float32_to_int32_round_to_zero float32_to_int32_round_to_zero_x86_64
#define float32_to_int64_round_to_zero float32_to_int64_round_to_zero_x86_64
#define float64_to_int16_round_to_zero float64_to_int16_round_to_zero_x86_64
#define float64_to_int32_round_to_zero float64_to_int32_round_to_zero_x86_64
#define float64_to_int64_round_to_zero float64_to_int64_round_to_zero_x86_64
#define float16_to_uint16_scalbn float16_to_uint16_scalbn_x86_64
#define float16_to_uint32_scalbn float16_to_uint32_scalbn_x86_64
#define float16_to_uint64_scalbn float16_to_uint64_scalbn_x86_64
#define float32_to_uint16_scalbn float32_to_uint16_scalbn_x86_64
#define float32_to_uint32_scalbn float32_to_uint32_scalbn_x86_64
#define float32_to_uint64_scalbn float32_to_uint64_scalbn_x86_64
#define float64_to_uint16_scalbn float64_to_uint16_scalbn_x86_64
#define float64_to_uint32_scalbn float64_to_uint32_scalbn_x86_64
#define float64_to_uint64_scalbn float64_to_uint64_scalbn_x86_64
#define float16_to_uint16 float16_to_uint16_x86_64
#define float16_to_uint32 float16_to_uint32_x86_64
#define float16_to_uint64 float16_to_uint64_x86_64
#define float32_to_uint16 float32_to_uint16_x86_64
#define float32_to_uint32 float32_to_uint32_x86_64
#define float32_to_uint64 float32_to_uint64_x86_64
#define float64_to_uint16 float64_to_uint16_x86_64
#define float64_to_uint32 float64_to_uint32_x86_64
#define float64_to_uint64 float64_to_uint64_x86_64
#define float16_to_uint16_round_to_zero float16_to_uint16_round_to_zero_x86_64
#define float16_to_uint32_round_to_zero float16_to_uint32_round_to_zero_x86_64
#define float16_to_uint64_round_to_zero float16_to_uint64_round_to_zero_x86_64
#define float32_to_uint16_round_to_zero float32_to_uint16_round_to_zero_x86_64
#define float32_to_uint32_round_to_zero float32_to_uint32_round_to_zero_x86_64
#define float32_to_uint64_round_to_zero float32_to_uint64_round_to_zero_x86_64
#define float64_to_uint16_round_to_zero float64_to_uint16_round_to_zero_x86_64
#define float64_to_uint32_round_to_zero float64_to_uint32_round_to_zero_x86_64
#define float64_to_uint64_round_to_zero float64_to_uint64_round_to_zero_x86_64
#define int64_to_float16_scalbn int64_to_float16_scalbn_x86_64
#define int32_to_float16_scalbn int32_to_float16_scalbn_x86_64
#define int16_to_float16_scalbn int16_to_float16_scalbn_x86_64
#define int64_to_float16 int64_to_float16_x86_64
#define int32_to_float16 int32_to_float16_x86_64
#define int16_to_float16 int16_to_float16_x86_64
#define int64_to_float32_scalbn int64_to_float32_scalbn_x86_64
#define int32_to_float32_scalbn int32_to_float32_scalbn_x86_64
#define int16_to_float32_scalbn int16_to_float32_scalbn_x86_64
#define int64_to_float32 int64_to_float32_x86_64
#define int32_to_float32 int32_to_float32_x86_64
#define int16_to_float32 int16_to_float32_x86_64
#define int64_to_float64_scalbn int64_to_float64_scalbn_x86_64
#define int32_to_float64_scalbn int32_to_float64_scalbn_x86_64
#define int16_to_float64_scalbn int16_to_float64_scalbn_x86_64
#define int64_to_float64 int64_to_float64_x86_64
#define int32_to_float64 int32_to_float64_x86_64
#define int16_to_float64 int16_to_float64_x86_64
#define uint64_to_float16_scalbn uint64_to_float16_scalbn_x86_64
#define uint32_to_float16_scalbn uint32_to_float16_scalbn_x86_64
#define uint16_to_float16_scalbn uint16_to_float16_scalbn_x86_64
#define uint64_to_float16 uint64_to_float16_x86_64
#define uint32_to_float16 uint32_to_float16_x86_64
#define uint16_to_float16 uint16_to_float16_x86_64
#define uint64_to_float32_scalbn uint64_to_float32_scalbn_x86_64
#define uint32_to_float32_scalbn uint32_to_float32_scalbn_x86_64
#define uint16_to_float32_scalbn uint16_to_float32_scalbn_x86_64
#define uint64_to_float32 uint64_to_float32_x86_64
#define uint32_to_float32 uint32_to_float32_x86_64
#define uint16_to_float32 uint16_to_float32_x86_64
#define uint64_to_float64_scalbn uint64_to_float64_scalbn_x86_64
#define uint32_to_float64_scalbn uint32_to_float64_scalbn_x86_64
#define uint16_to_float64_scalbn uint16_to_float64_scalbn_x86_64
#define uint64_to_float64 uint64_to_float64_x86_64
#define uint32_to_float64 uint32_to_float64_x86_64
#define uint16_to_float64 uint16_to_float64_x86_64
#define float16_min float16_min_x86_64
#define float16_minnum float16_minnum_x86_64
#define float16_minnummag float16_minnummag_x86_64
#define float16_max float16_max_x86_64
#define float16_maxnum float16_maxnum_x86_64
#define float16_maxnummag float16_maxnummag_x86_64
#define float32_min float32_min_x86_64
#define float32_minnum float32_minnum_x86_64
#define float32_minnummag float32_minnummag_x86_64
#define float32_max float32_max_x86_64
#define float32_maxnum float32_maxnum_x86_64
#define float32_maxnummag float32_maxnummag_x86_64
#define float64_min float64_min_x86_64
#define float64_minnum float64_minnum_x86_64
#define float64_minnummag float64_minnummag_x86_64
#define float64_max float64_max_x86_64
#define float64_maxnum float64_maxnum_x86_64
#define float64_maxnummag float64_maxnummag_x86_64
#define float16_compare float16_compare_x86_64
#define float16_compare_quiet float16_compare_quiet_x86_64
#define float32_compare float32_compare_x86_64
#define float32_compare_quiet float32_compare_quiet_x86_64
#define float64_compare float64_compare_x86_64
#define float64_compare_quiet float64_compare_quiet_x86_64
#define float16_scalbn float16_scalbn_x86_64
#define float32_scalbn float32_scalbn_x86_64
#define float64_scalbn float64_scalbn_x86_64
#define float16_sqrt float16_sqrt_x86_64
#define float32_sqrt float32_sqrt_x86_64
#define float64_sqrt float64_sqrt_x86_64
#define float16_default_nan float16_default_nan_x86_64
#define float32_default_nan float32_default_nan_x86_64
#define float64_default_nan float64_default_nan_x86_64
#define float128_default_nan float128_default_nan_x86_64
#define float16_silence_nan float16_silence_nan_x86_64
#define float32_silence_nan float32_silence_nan_x86_64
#define float64_silence_nan float64_silence_nan_x86_64
#define float16_squash_input_denormal float16_squash_input_denormal_x86_64
#define float32_squash_input_denormal float32_squash_input_denormal_x86_64
#define float64_squash_input_denormal float64_squash_input_denormal_x86_64
#define normalizeFloatx80Subnormal normalizeFloatx80Subnormal_x86_64
#define roundAndPackFloatx80 roundAndPackFloatx80_x86_64
#define normalizeRoundAndPackFloatx80 normalizeRoundAndPackFloatx80_x86_64
#define int32_to_floatx80 int32_to_floatx80_x86_64
#define int32_to_float128 int32_to_float128_x86_64
#define int64_to_floatx80 int64_to_floatx80_x86_64
#define int64_to_float128 int64_to_float128_x86_64
#define uint64_to_float128 uint64_to_float128_x86_64
#define float32_to_floatx80 float32_to_floatx80_x86_64
#define float32_to_float128 float32_to_float128_x86_64
#define float32_rem float32_rem_x86_64
#define float32_exp2 float32_exp2_x86_64
#define float32_log2 float32_log2_x86_64
#define float32_eq float32_eq_x86_64
#define float32_le float32_le_x86_64
#define float32_lt float32_lt_x86_64
#define float32_unordered float32_unordered_x86_64
#define float32_eq_quiet float32_eq_quiet_x86_64
#define float32_le_quiet float32_le_quiet_x86_64
#define float32_lt_quiet float32_lt_quiet_x86_64
#define float32_unordered_quiet float32_unordered_quiet_x86_64
#define float64_to_floatx80 float64_to_floatx80_x86_64
#define float64_to_float128 float64_to_float128_x86_64
#define float64_rem float64_rem_x86_64
#define float64_log2 float64_log2_x86_64
#define float64_eq float64_eq_x86_64
#define float64_le float64_le_x86_64
#define float64_lt float64_lt_x86_64
#define float64_unordered float64_unordered_x86_64
#define float64_eq_quiet float64_eq_quiet_x86_64
#define float64_le_quiet float64_le_quiet_x86_64
#define float64_lt_quiet float64_lt_quiet_x86_64
#define float64_unordered_quiet float64_unordered_quiet_x86_64
#define floatx80_to_int32 floatx80_to_int32_x86_64
#define floatx80_to_int32_round_to_zero floatx80_to_int32_round_to_zero_x86_64
#define floatx80_to_int64 floatx80_to_int64_x86_64
#define floatx80_to_int64_round_to_zero floatx80_to_int64_round_to_zero_x86_64
#define floatx80_to_float32 floatx80_to_float32_x86_64
#define floatx80_to_float64 floatx80_to_float64_x86_64
#define floatx80_to_float128 floatx80_to_float128_x86_64
#define floatx80_round floatx80_round_x86_64
#define floatx80_round_to_int floatx80_round_to_int_x86_64
#define floatx80_add floatx80_add_x86_64
#define floatx80_sub floatx80_sub_x86_64
#define floatx80_mul floatx80_mul_x86_64
#define floatx80_div floatx80_div_x86_64
#define floatx80_rem floatx80_rem_x86_64
#define floatx80_sqrt floatx80_sqrt_x86_64
#define floatx80_eq floatx80_eq_x86_64
#define floatx80_le floatx80_le_x86_64
#define floatx80_lt floatx80_lt_x86_64
#define floatx80_unordered floatx80_unordered_x86_64
#define floatx80_eq_quiet floatx80_eq_quiet_x86_64
#define floatx80_le_quiet floatx80_le_quiet_x86_64
#define floatx80_lt_quiet floatx80_lt_quiet_x86_64
#define floatx80_unordered_quiet floatx80_unordered_quiet_x86_64
#define float128_to_int32 float128_to_int32_x86_64
#define float128_to_int32_round_to_zero float128_to_int32_round_to_zero_x86_64
#define float128_to_int64 float128_to_int64_x86_64
#define float128_to_int64_round_to_zero float128_to_int64_round_to_zero_x86_64
#define float128_to_uint64 float128_to_uint64_x86_64
#define float128_to_uint64_round_to_zero float128_to_uint64_round_to_zero_x86_64
#define float128_to_uint32_round_to_zero float128_to_uint32_round_to_zero_x86_64
#define float128_to_uint32 float128_to_uint32_x86_64
#define float128_to_float32 float128_to_float32_x86_64
#define float128_to_float64 float128_to_float64_x86_64
#define float128_to_floatx80 float128_to_floatx80_x86_64
#define float128_round_to_int float128_round_to_int_x86_64
#define float128_add float128_add_x86_64
#define float128_sub float128_sub_x86_64
#define float128_mul float128_mul_x86_64
#define float128_div float128_div_x86_64
#define float128_rem float128_rem_x86_64
#define float128_sqrt float128_sqrt_x86_64
#define float128_eq float128_eq_x86_64
#define float128_le float128_le_x86_64
#define float128_lt float128_lt_x86_64
#define float128_unordered float128_unordered_x86_64
#define float128_eq_quiet float128_eq_quiet_x86_64
#define float128_le_quiet float128_le_quiet_x86_64
#define float128_lt_quiet float128_lt_quiet_x86_64
#define float128_unordered_quiet float128_unordered_quiet_x86_64
#define floatx80_compare floatx80_compare_x86_64
#define floatx80_compare_quiet floatx80_compare_quiet_x86_64
#define float128_compare float128_compare_x86_64
#define float128_compare_quiet float128_compare_quiet_x86_64
#define floatx80_scalbn floatx80_scalbn_x86_64
#define float128_scalbn float128_scalbn_x86_64
#define softfloat_init softfloat_init_x86_64
#define tcg_optimize tcg_optimize_x86_64
#define gen_new_label gen_new_label_x86_64
#define tcg_can_emit_vec_op tcg_can_emit_vec_op_x86_64
#define tcg_expand_vec_op tcg_expand_vec_op_x86_64
#define tcg_register_jit tcg_register_jit_x86_64
#define tcg_tb_insert tcg_tb_insert_x86_64
#define tcg_tb_remove tcg_tb_remove_x86_64
#define tcg_tb_lookup tcg_tb_lookup_x86_64
#define tcg_tb_foreach tcg_tb_foreach_x86_64
#define tcg_nb_tbs tcg_nb_tbs_x86_64
#define tcg_region_reset_all tcg_region_reset_all_x86_64
#define tcg_region_init tcg_region_init_x86_64
#define tcg_code_size tcg_code_size_x86_64
#define tcg_code_capacity tcg_code_capacity_x86_64
#define tcg_tb_phys_invalidate_count tcg_tb_phys_invalidate_count_x86_64
#define tcg_malloc_internal tcg_malloc_internal_x86_64
#define tcg_pool_reset tcg_pool_reset_x86_64
#define tcg_context_init tcg_context_init_x86_64
#define tcg_tb_alloc tcg_tb_alloc_x86_64
#define tcg_prologue_init tcg_prologue_init_x86_64
#define tcg_func_start tcg_func_start_x86_64
#define tcg_set_frame tcg_set_frame_x86_64
#define tcg_global_mem_new_internal tcg_global_mem_new_internal_x86_64
#define tcg_temp_new_internal tcg_temp_new_internal_x86_64
#define tcg_temp_new_vec tcg_temp_new_vec_x86_64
#define tcg_temp_new_vec_matching tcg_temp_new_vec_matching_x86_64
#define tcg_temp_free_internal tcg_temp_free_internal_x86_64
#define tcg_const_i32 tcg_const_i32_x86_64
#define tcg_const_i64 tcg_const_i64_x86_64
#define tcg_const_local_i32 tcg_const_local_i32_x86_64
#define tcg_const_local_i64 tcg_const_local_i64_x86_64
#define tcg_op_supported tcg_op_supported_x86_64
#define tcg_gen_callN tcg_gen_callN_x86_64
#define tcg_op_remove tcg_op_remove_x86_64
#define tcg_emit_op tcg_emit_op_x86_64
#define tcg_op_insert_before tcg_op_insert_before_x86_64
#define tcg_op_insert_after tcg_op_insert_after_x86_64
#define tcg_cpu_exec_time tcg_cpu_exec_time_x86_64
#define tcg_gen_code tcg_gen_code_x86_64
#define tcg_gen_op1 tcg_gen_op1_x86_64
#define tcg_gen_op2 tcg_gen_op2_x86_64
#define tcg_gen_op3 tcg_gen_op3_x86_64
#define tcg_gen_op4 tcg_gen_op4_x86_64
#define tcg_gen_op5 tcg_gen_op5_x86_64
#define tcg_gen_op6 tcg_gen_op6_x86_64
#define tcg_gen_mb tcg_gen_mb_x86_64
#define tcg_gen_addi_i32 tcg_gen_addi_i32_x86_64
#define tcg_gen_subfi_i32 tcg_gen_subfi_i32_x86_64
#define tcg_gen_subi_i32 tcg_gen_subi_i32_x86_64
#define tcg_gen_andi_i32 tcg_gen_andi_i32_x86_64
#define tcg_gen_ori_i32 tcg_gen_ori_i32_x86_64
#define tcg_gen_xori_i32 tcg_gen_xori_i32_x86_64
#define tcg_gen_shli_i32 tcg_gen_shli_i32_x86_64
#define tcg_gen_shri_i32 tcg_gen_shri_i32_x86_64
#define tcg_gen_sari_i32 tcg_gen_sari_i32_x86_64
#define tcg_gen_brcond_i32 tcg_gen_brcond_i32_x86_64
#define tcg_gen_brcondi_i32 tcg_gen_brcondi_i32_x86_64
#define tcg_gen_setcond_i32 tcg_gen_setcond_i32_x86_64
#define tcg_gen_setcondi_i32 tcg_gen_setcondi_i32_x86_64
#define tcg_gen_muli_i32 tcg_gen_muli_i32_x86_64
#define tcg_gen_div_i32 tcg_gen_div_i32_x86_64
#define tcg_gen_rem_i32 tcg_gen_rem_i32_x86_64
#define tcg_gen_divu_i32 tcg_gen_divu_i32_x86_64
#define tcg_gen_remu_i32 tcg_gen_remu_i32_x86_64
#define tcg_gen_andc_i32 tcg_gen_andc_i32_x86_64
#define tcg_gen_eqv_i32 tcg_gen_eqv_i32_x86_64
#define tcg_gen_nand_i32 tcg_gen_nand_i32_x86_64
#define tcg_gen_nor_i32 tcg_gen_nor_i32_x86_64
#define tcg_gen_orc_i32 tcg_gen_orc_i32_x86_64
#define tcg_gen_clz_i32 tcg_gen_clz_i32_x86_64
#define tcg_gen_clzi_i32 tcg_gen_clzi_i32_x86_64
#define tcg_gen_ctz_i32 tcg_gen_ctz_i32_x86_64
#define tcg_gen_ctzi_i32 tcg_gen_ctzi_i32_x86_64
#define tcg_gen_clrsb_i32 tcg_gen_clrsb_i32_x86_64
#define tcg_gen_ctpop_i32 tcg_gen_ctpop_i32_x86_64
#define tcg_gen_rotl_i32 tcg_gen_rotl_i32_x86_64
#define tcg_gen_rotli_i32 tcg_gen_rotli_i32_x86_64
#define tcg_gen_rotr_i32 tcg_gen_rotr_i32_x86_64
#define tcg_gen_rotri_i32 tcg_gen_rotri_i32_x86_64
#define tcg_gen_deposit_i32 tcg_gen_deposit_i32_x86_64
#define tcg_gen_deposit_z_i32 tcg_gen_deposit_z_i32_x86_64
#define tcg_gen_extract_i32 tcg_gen_extract_i32_x86_64
#define tcg_gen_sextract_i32 tcg_gen_sextract_i32_x86_64
#define tcg_gen_extract2_i32 tcg_gen_extract2_i32_x86_64
#define tcg_gen_movcond_i32 tcg_gen_movcond_i32_x86_64
#define tcg_gen_add2_i32 tcg_gen_add2_i32_x86_64
#define tcg_gen_sub2_i32 tcg_gen_sub2_i32_x86_64
#define tcg_gen_mulu2_i32 tcg_gen_mulu2_i32_x86_64
#define tcg_gen_muls2_i32 tcg_gen_muls2_i32_x86_64
#define tcg_gen_mulsu2_i32 tcg_gen_mulsu2_i32_x86_64
#define tcg_gen_ext8s_i32 tcg_gen_ext8s_i32_x86_64
#define tcg_gen_ext16s_i32 tcg_gen_ext16s_i32_x86_64
#define tcg_gen_ext8u_i32 tcg_gen_ext8u_i32_x86_64
#define tcg_gen_ext16u_i32 tcg_gen_ext16u_i32_x86_64
#define tcg_gen_bswap16_i32 tcg_gen_bswap16_i32_x86_64
#define tcg_gen_bswap32_i32 tcg_gen_bswap32_i32_x86_64
#define tcg_gen_smin_i32 tcg_gen_smin_i32_x86_64
#define tcg_gen_umin_i32 tcg_gen_umin_i32_x86_64
#define tcg_gen_smax_i32 tcg_gen_smax_i32_x86_64
#define tcg_gen_umax_i32 tcg_gen_umax_i32_x86_64
#define tcg_gen_abs_i32 tcg_gen_abs_i32_x86_64
#define tcg_gen_addi_i64 tcg_gen_addi_i64_x86_64
#define tcg_gen_subfi_i64 tcg_gen_subfi_i64_x86_64
#define tcg_gen_subi_i64 tcg_gen_subi_i64_x86_64
#define tcg_gen_andi_i64 tcg_gen_andi_i64_x86_64
#define tcg_gen_ori_i64 tcg_gen_ori_i64_x86_64
#define tcg_gen_xori_i64 tcg_gen_xori_i64_x86_64
#define tcg_gen_shli_i64 tcg_gen_shli_i64_x86_64
#define tcg_gen_shri_i64 tcg_gen_shri_i64_x86_64
#define tcg_gen_sari_i64 tcg_gen_sari_i64_x86_64
#define tcg_gen_brcond_i64 tcg_gen_brcond_i64_x86_64
#define tcg_gen_brcondi_i64 tcg_gen_brcondi_i64_x86_64
#define tcg_gen_setcond_i64 tcg_gen_setcond_i64_x86_64
#define tcg_gen_setcondi_i64 tcg_gen_setcondi_i64_x86_64
#define tcg_gen_muli_i64 tcg_gen_muli_i64_x86_64
#define tcg_gen_div_i64 tcg_gen_div_i64_x86_64
#define tcg_gen_rem_i64 tcg_gen_rem_i64_x86_64
#define tcg_gen_divu_i64 tcg_gen_divu_i64_x86_64
#define tcg_gen_remu_i64 tcg_gen_remu_i64_x86_64
#define tcg_gen_ext8s_i64 tcg_gen_ext8s_i64_x86_64
#define tcg_gen_ext16s_i64 tcg_gen_ext16s_i64_x86_64
#define tcg_gen_ext32s_i64 tcg_gen_ext32s_i64_x86_64
#define tcg_gen_ext8u_i64 tcg_gen_ext8u_i64_x86_64
#define tcg_gen_ext16u_i64 tcg_gen_ext16u_i64_x86_64
#define tcg_gen_ext32u_i64 tcg_gen_ext32u_i64_x86_64
#define tcg_gen_bswap16_i64 tcg_gen_bswap16_i64_x86_64
#define tcg_gen_bswap32_i64 tcg_gen_bswap32_i64_x86_64
#define tcg_gen_bswap64_i64 tcg_gen_bswap64_i64_x86_64
#define tcg_gen_not_i64 tcg_gen_not_i64_x86_64
#define tcg_gen_andc_i64 tcg_gen_andc_i64_x86_64
#define tcg_gen_eqv_i64 tcg_gen_eqv_i64_x86_64
#define tcg_gen_nand_i64 tcg_gen_nand_i64_x86_64
#define tcg_gen_nor_i64 tcg_gen_nor_i64_x86_64
#define tcg_gen_orc_i64 tcg_gen_orc_i64_x86_64
#define tcg_gen_clz_i64 tcg_gen_clz_i64_x86_64
#define tcg_gen_clzi_i64 tcg_gen_clzi_i64_x86_64
#define tcg_gen_ctz_i64 tcg_gen_ctz_i64_x86_64
#define tcg_gen_ctzi_i64 tcg_gen_ctzi_i64_x86_64
#define tcg_gen_clrsb_i64 tcg_gen_clrsb_i64_x86_64
#define tcg_gen_ctpop_i64 tcg_gen_ctpop_i64_x86_64
#define tcg_gen_rotl_i64 tcg_gen_rotl_i64_x86_64
#define tcg_gen_rotli_i64 tcg_gen_rotli_i64_x86_64
#define tcg_gen_rotr_i64 tcg_gen_rotr_i64_x86_64
#define tcg_gen_rotri_i64 tcg_gen_rotri_i64_x86_64
#define tcg_gen_deposit_i64 tcg_gen_deposit_i64_x86_64
#define tcg_gen_deposit_z_i64 tcg_gen_deposit_z_i64_x86_64
#define tcg_gen_extract_i64 tcg_gen_extract_i64_x86_64
#define tcg_gen_sextract_i64 tcg_gen_sextract_i64_x86_64
#define tcg_gen_extract2_i64 tcg_gen_extract2_i64_x86_64
#define tcg_gen_movcond_i64 tcg_gen_movcond_i64_x86_64
#define tcg_gen_add2_i64 tcg_gen_add2_i64_x86_64
#define tcg_gen_sub2_i64 tcg_gen_sub2_i64_x86_64
#define tcg_gen_mulu2_i64 tcg_gen_mulu2_i64_x86_64
#define tcg_gen_muls2_i64 tcg_gen_muls2_i64_x86_64
#define tcg_gen_mulsu2_i64 tcg_gen_mulsu2_i64_x86_64
#define tcg_gen_smin_i64 tcg_gen_smin_i64_x86_64
#define tcg_gen_umin_i64 tcg_gen_umin_i64_x86_64
#define tcg_gen_smax_i64 tcg_gen_smax_i64_x86_64
#define tcg_gen_umax_i64 tcg_gen_umax_i64_x86_64
#define tcg_gen_abs_i64 tcg_gen_abs_i64_x86_64
#define tcg_gen_extrl_i64_i32 tcg_gen_extrl_i64_i32_x86_64
#define tcg_gen_extrh_i64_i32 tcg_gen_extrh_i64_i32_x86_64
#define tcg_gen_extu_i32_i64 tcg_gen_extu_i32_i64_x86_64
#define tcg_gen_ext_i32_i64 tcg_gen_ext_i32_i64_x86_64
#define tcg_gen_concat_i32_i64 tcg_gen_concat_i32_i64_x86_64
#define tcg_gen_extr_i64_i32 tcg_gen_extr_i64_i32_x86_64
#define tcg_gen_extr32_i64 tcg_gen_extr32_i64_x86_64
#define tcg_gen_exit_tb tcg_gen_exit_tb_x86_64
#define tcg_gen_goto_tb tcg_gen_goto_tb_x86_64
#define tcg_gen_lookup_and_goto_ptr tcg_gen_lookup_and_goto_ptr_x86_64
#define check_exit_request check_exit_request_x86_64
#define tcg_gen_qemu_ld_i32 tcg_gen_qemu_ld_i32_x86_64
#define tcg_gen_qemu_st_i32 tcg_gen_qemu_st_i32_x86_64
#define tcg_gen_qemu_ld_i64 tcg_gen_qemu_ld_i64_x86_64
#define tcg_gen_qemu_st_i64 tcg_gen_qemu_st_i64_x86_64
#define tcg_gen_atomic_cmpxchg_i32 tcg_gen_atomic_cmpxchg_i32_x86_64
#define tcg_gen_atomic_cmpxchg_i64 tcg_gen_atomic_cmpxchg_i64_x86_64
#define tcg_gen_atomic_fetch_add_i32 tcg_gen_atomic_fetch_add_i32_x86_64
#define tcg_gen_atomic_fetch_add_i64 tcg_gen_atomic_fetch_add_i64_x86_64
#define tcg_gen_atomic_fetch_and_i32 tcg_gen_atomic_fetch_and_i32_x86_64
#define tcg_gen_atomic_fetch_and_i64 tcg_gen_atomic_fetch_and_i64_x86_64
#define tcg_gen_atomic_fetch_or_i32 tcg_gen_atomic_fetch_or_i32_x86_64
#define tcg_gen_atomic_fetch_or_i64 tcg_gen_atomic_fetch_or_i64_x86_64
#define tcg_gen_atomic_fetch_xor_i32 tcg_gen_atomic_fetch_xor_i32_x86_64
#define tcg_gen_atomic_fetch_xor_i64 tcg_gen_atomic_fetch_xor_i64_x86_64
#define tcg_gen_atomic_fetch_smin_i32 tcg_gen_atomic_fetch_smin_i32_x86_64
#define tcg_gen_atomic_fetch_smin_i64 tcg_gen_atomic_fetch_smin_i64_x86_64
#define tcg_gen_atomic_fetch_umin_i32 tcg_gen_atomic_fetch_umin_i32_x86_64
#define tcg_gen_atomic_fetch_umin_i64 tcg_gen_atomic_fetch_umin_i64_x86_64
#define tcg_gen_atomic_fetch_smax_i32 tcg_gen_atomic_fetch_smax_i32_x86_64
#define tcg_gen_atomic_fetch_smax_i64 tcg_gen_atomic_fetch_smax_i64_x86_64
#define tcg_gen_atomic_fetch_umax_i32 tcg_gen_atomic_fetch_umax_i32_x86_64
#define tcg_gen_atomic_fetch_umax_i64 tcg_gen_atomic_fetch_umax_i64_x86_64
#define tcg_gen_atomic_add_fetch_i32 tcg_gen_atomic_add_fetch_i32_x86_64
#define tcg_gen_atomic_add_fetch_i64 tcg_gen_atomic_add_fetch_i64_x86_64
#define tcg_gen_atomic_and_fetch_i32 tcg_gen_atomic_and_fetch_i32_x86_64
#define tcg_gen_atomic_and_fetch_i64 tcg_gen_atomic_and_fetch_i64_x86_64
#define tcg_gen_atomic_or_fetch_i32 tcg_gen_atomic_or_fetch_i32_x86_64
#define tcg_gen_atomic_or_fetch_i64 tcg_gen_atomic_or_fetch_i64_x86_64
#define tcg_gen_atomic_xor_fetch_i32 tcg_gen_atomic_xor_fetch_i32_x86_64
#define tcg_gen_atomic_xor_fetch_i64 tcg_gen_atomic_xor_fetch_i64_x86_64
#define tcg_gen_atomic_smin_fetch_i32 tcg_gen_atomic_smin_fetch_i32_x86_64
#define tcg_gen_atomic_smin_fetch_i64 tcg_gen_atomic_smin_fetch_i64_x86_64
#define tcg_gen_atomic_umin_fetch_i32 tcg_gen_atomic_umin_fetch_i32_x86_64
#define tcg_gen_atomic_umin_fetch_i64 tcg_gen_atomic_umin_fetch_i64_x86_64
#define tcg_gen_atomic_smax_fetch_i32 tcg_gen_atomic_smax_fetch_i32_x86_64
#define tcg_gen_atomic_smax_fetch_i64 tcg_gen_atomic_smax_fetch_i64_x86_64
#define tcg_gen_atomic_umax_fetch_i32 tcg_gen_atomic_umax_fetch_i32_x86_64
#define tcg_gen_atomic_umax_fetch_i64 tcg_gen_atomic_umax_fetch_i64_x86_64
#define tcg_gen_atomic_xchg_i32 tcg_gen_atomic_xchg_i32_x86_64
#define tcg_gen_atomic_xchg_i64 tcg_gen_atomic_xchg_i64_x86_64
#define simd_desc simd_desc_x86_64
#define tcg_gen_gvec_2_ool tcg_gen_gvec_2_ool_x86_64
#define tcg_gen_gvec_2i_ool tcg_gen_gvec_2i_ool_x86_64
#define tcg_gen_gvec_3_ool tcg_gen_gvec_3_ool_x86_64
#define tcg_gen_gvec_4_ool tcg_gen_gvec_4_ool_x86_64
#define tcg_gen_gvec_5_ool tcg_gen_gvec_5_ool_x86_64
#define tcg_gen_gvec_2_ptr tcg_gen_gvec_2_ptr_x86_64
#define tcg_gen_gvec_3_ptr tcg_gen_gvec_3_ptr_x86_64
#define tcg_gen_gvec_4_ptr tcg_gen_gvec_4_ptr_x86_64
#define tcg_gen_gvec_5_ptr tcg_gen_gvec_5_ptr_x86_64
#define tcg_gen_gvec_2 tcg_gen_gvec_2_x86_64
#define tcg_gen_gvec_2i tcg_gen_gvec_2i_x86_64
#define tcg_gen_gvec_2s tcg_gen_gvec_2s_x86_64
#define tcg_gen_gvec_3 tcg_gen_gvec_3_x86_64
#define tcg_gen_gvec_3i tcg_gen_gvec_3i_x86_64
#define tcg_gen_gvec_4 tcg_gen_gvec_4_x86_64
#define tcg_gen_gvec_mov tcg_gen_gvec_mov_x86_64
#define tcg_gen_gvec_dup_i32 tcg_gen_gvec_dup_i32_x86_64
#define tcg_gen_gvec_dup_i64 tcg_gen_gvec_dup_i64_x86_64
#define tcg_gen_gvec_dup_mem tcg_gen_gvec_dup_mem_x86_64
#define tcg_gen_gvec_dup64i tcg_gen_gvec_dup64i_x86_64
#define tcg_gen_gvec_dup32i tcg_gen_gvec_dup32i_x86_64
#define tcg_gen_gvec_dup16i tcg_gen_gvec_dup16i_x86_64
#define tcg_gen_gvec_dup8i tcg_gen_gvec_dup8i_x86_64
#define tcg_gen_gvec_not tcg_gen_gvec_not_x86_64
#define tcg_gen_vec_add8_i64 tcg_gen_vec_add8_i64_x86_64
#define tcg_gen_vec_add16_i64 tcg_gen_vec_add16_i64_x86_64
#define tcg_gen_vec_add32_i64 tcg_gen_vec_add32_i64_x86_64
#define tcg_gen_gvec_add tcg_gen_gvec_add_x86_64
#define tcg_gen_gvec_adds tcg_gen_gvec_adds_x86_64
#define tcg_gen_gvec_addi tcg_gen_gvec_addi_x86_64
#define tcg_gen_gvec_subs tcg_gen_gvec_subs_x86_64
#define tcg_gen_vec_sub8_i64 tcg_gen_vec_sub8_i64_x86_64
#define tcg_gen_vec_sub16_i64 tcg_gen_vec_sub16_i64_x86_64
#define tcg_gen_vec_sub32_i64 tcg_gen_vec_sub32_i64_x86_64
#define tcg_gen_gvec_sub tcg_gen_gvec_sub_x86_64
#define tcg_gen_gvec_mul tcg_gen_gvec_mul_x86_64
#define tcg_gen_gvec_muls tcg_gen_gvec_muls_x86_64
#define tcg_gen_gvec_muli tcg_gen_gvec_muli_x86_64
#define tcg_gen_gvec_ssadd tcg_gen_gvec_ssadd_x86_64
#define tcg_gen_gvec_sssub tcg_gen_gvec_sssub_x86_64
#define tcg_gen_gvec_usadd tcg_gen_gvec_usadd_x86_64
#define tcg_gen_gvec_ussub tcg_gen_gvec_ussub_x86_64
#define tcg_gen_gvec_smin tcg_gen_gvec_smin_x86_64
#define tcg_gen_gvec_umin tcg_gen_gvec_umin_x86_64
#define tcg_gen_gvec_smax tcg_gen_gvec_smax_x86_64
#define tcg_gen_gvec_umax tcg_gen_gvec_umax_x86_64
#define tcg_gen_vec_neg8_i64 tcg_gen_vec_neg8_i64_x86_64
#define tcg_gen_vec_neg16_i64 tcg_gen_vec_neg16_i64_x86_64
#define tcg_gen_vec_neg32_i64 tcg_gen_vec_neg32_i64_x86_64
#define tcg_gen_gvec_neg tcg_gen_gvec_neg_x86_64
#define tcg_gen_gvec_abs tcg_gen_gvec_abs_x86_64
#define tcg_gen_gvec_and tcg_gen_gvec_and_x86_64
#define tcg_gen_gvec_or tcg_gen_gvec_or_x86_64
#define tcg_gen_gvec_xor tcg_gen_gvec_xor_x86_64
#define tcg_gen_gvec_andc tcg_gen_gvec_andc_x86_64
#define tcg_gen_gvec_orc tcg_gen_gvec_orc_x86_64
#define tcg_gen_gvec_nand tcg_gen_gvec_nand_x86_64
#define tcg_gen_gvec_nor tcg_gen_gvec_nor_x86_64
#define tcg_gen_gvec_eqv tcg_gen_gvec_eqv_x86_64
#define tcg_gen_gvec_ands tcg_gen_gvec_ands_x86_64
#define tcg_gen_gvec_andi tcg_gen_gvec_andi_x86_64
#define tcg_gen_gvec_xors tcg_gen_gvec_xors_x86_64
#define tcg_gen_gvec_xori tcg_gen_gvec_xori_x86_64
#define tcg_gen_gvec_ors tcg_gen_gvec_ors_x86_64
#define tcg_gen_gvec_ori tcg_gen_gvec_ori_x86_64
#define tcg_gen_vec_shl8i_i64 tcg_gen_vec_shl8i_i64_x86_64
#define tcg_gen_vec_shl16i_i64 tcg_gen_vec_shl16i_i64_x86_64
#define tcg_gen_gvec_shli tcg_gen_gvec_shli_x86_64
#define tcg_gen_vec_shr8i_i64 tcg_gen_vec_shr8i_i64_x86_64
#define tcg_gen_vec_shr16i_i64 tcg_gen_vec_shr16i_i64_x86_64
#define tcg_gen_gvec_shri tcg_gen_gvec_shri_x86_64
#define tcg_gen_vec_sar8i_i64 tcg_gen_vec_sar8i_i64_x86_64
#define tcg_gen_vec_sar16i_i64 tcg_gen_vec_sar16i_i64_x86_64
#define tcg_gen_gvec_sari tcg_gen_gvec_sari_x86_64
#define tcg_gen_gvec_shls tcg_gen_gvec_shls_x86_64
#define tcg_gen_gvec_shrs tcg_gen_gvec_shrs_x86_64
#define tcg_gen_gvec_sars tcg_gen_gvec_sars_x86_64
#define tcg_gen_gvec_shlv tcg_gen_gvec_shlv_x86_64
#define tcg_gen_gvec_shrv tcg_gen_gvec_shrv_x86_64
#define tcg_gen_gvec_sarv tcg_gen_gvec_sarv_x86_64
#define tcg_gen_gvec_cmp tcg_gen_gvec_cmp_x86_64
#define tcg_gen_gvec_bitsel tcg_gen_gvec_bitsel_x86_64
#define tcg_can_emit_vecop_list tcg_can_emit_vecop_list_x86_64
#define vec_gen_2 vec_gen_2_x86_64
#define vec_gen_3 vec_gen_3_x86_64
#define vec_gen_4 vec_gen_4_x86_64
#define tcg_gen_mov_vec tcg_gen_mov_vec_x86_64
#define tcg_const_zeros_vec tcg_const_zeros_vec_x86_64
#define tcg_const_ones_vec tcg_const_ones_vec_x86_64
#define tcg_const_zeros_vec_matching tcg_const_zeros_vec_matching_x86_64
#define tcg_const_ones_vec_matching tcg_const_ones_vec_matching_x86_64
#define tcg_gen_dup64i_vec tcg_gen_dup64i_vec_x86_64
#define tcg_gen_dup32i_vec tcg_gen_dup32i_vec_x86_64
#define tcg_gen_dup16i_vec tcg_gen_dup16i_vec_x86_64
#define tcg_gen_dup8i_vec tcg_gen_dup8i_vec_x86_64
#define tcg_gen_dupi_vec tcg_gen_dupi_vec_x86_64
#define tcg_gen_dup_i64_vec tcg_gen_dup_i64_vec_x86_64
#define tcg_gen_dup_i32_vec tcg_gen_dup_i32_vec_x86_64
#define tcg_gen_dup_mem_vec tcg_gen_dup_mem_vec_x86_64
#define tcg_gen_ld_vec tcg_gen_ld_vec_x86_64
#define tcg_gen_st_vec tcg_gen_st_vec_x86_64
#define tcg_gen_stl_vec tcg_gen_stl_vec_x86_64
#define tcg_gen_and_vec tcg_gen_and_vec_x86_64
#define tcg_gen_or_vec tcg_gen_or_vec_x86_64
#define tcg_gen_xor_vec tcg_gen_xor_vec_x86_64
#define tcg_gen_andc_vec tcg_gen_andc_vec_x86_64
#define tcg_gen_orc_vec tcg_gen_orc_vec_x86_64
#define tcg_gen_nand_vec tcg_gen_nand_vec_x86_64
#define tcg_gen_nor_vec tcg_gen_nor_vec_x86_64
#define tcg_gen_eqv_vec tcg_gen_eqv_vec_x86_64
#define tcg_gen_not_vec tcg_gen_not_vec_x86_64
#define tcg_gen_neg_vec tcg_gen_neg_vec_x86_64
#define tcg_gen_abs_vec tcg_gen_abs_vec_x86_64
#define tcg_gen_shli_vec tcg_gen_shli_vec_x86_64
#define tcg_gen_shri_vec tcg_gen_shri_vec_x86_64
#define tcg_gen_sari_vec tcg_gen_sari_vec_x86_64
#define tcg_gen_cmp_vec tcg_gen_cmp_vec_x86_64
#define tcg_gen_add_vec tcg_gen_add_vec_x86_64
#define tcg_gen_sub_vec tcg_gen_sub_vec_x86_64
#define tcg_gen_mul_vec tcg_gen_mul_vec_x86_64
#define tcg_gen_ssadd_vec tcg_gen_ssadd_vec_x86_64
#define tcg_gen_usadd_vec tcg_gen_usadd_vec_x86_64
#define tcg_gen_sssub_vec tcg_gen_sssub_vec_x86_64
#define tcg_gen_ussub_vec tcg_gen_ussub_vec_x86_64
#define tcg_gen_smin_vec tcg_gen_smin_vec_x86_64
#define tcg_gen_umin_vec tcg_gen_umin_vec_x86_64
#define tcg_gen_smax_vec tcg_gen_smax_vec_x86_64
#define tcg_gen_umax_vec tcg_gen_umax_vec_x86_64
#define tcg_gen_shlv_vec tcg_gen_shlv_vec_x86_64
#define tcg_gen_shrv_vec tcg_gen_shrv_vec_x86_64
#define tcg_gen_sarv_vec tcg_gen_sarv_vec_x86_64
#define tcg_gen_shls_vec tcg_gen_shls_vec_x86_64
#define tcg_gen_shrs_vec tcg_gen_shrs_vec_x86_64
#define tcg_gen_sars_vec tcg_gen_sars_vec_x86_64
#define tcg_gen_bitsel_vec tcg_gen_bitsel_vec_x86_64
#define tcg_gen_cmpsel_vec tcg_gen_cmpsel_vec_x86_64
#define tb_htable_lookup tb_htable_lookup_x86_64
#define tb_set_jmp_target tb_set_jmp_target_x86_64
#define cpu_exec cpu_exec_x86_64
#define cpu_loop_exit_noexc cpu_loop_exit_noexc_x86_64
#define cpu_reloading_memory_map cpu_reloading_memory_map_x86_64
#define cpu_loop_exit cpu_loop_exit_x86_64
#define cpu_loop_exit_restore cpu_loop_exit_restore_x86_64
#define cpu_loop_exit_atomic cpu_loop_exit_atomic_x86_64
#define tlb_init tlb_init_x86_64
#define tlb_flush_by_mmuidx tlb_flush_by_mmuidx_x86_64
#define tlb_flush tlb_flush_x86_64
#define tlb_flush_by_mmuidx_all_cpus tlb_flush_by_mmuidx_all_cpus_x86_64
#define tlb_flush_all_cpus tlb_flush_all_cpus_x86_64
#define tlb_flush_by_mmuidx_all_cpus_synced tlb_flush_by_mmuidx_all_cpus_synced_x86_64
#define tlb_flush_all_cpus_synced tlb_flush_all_cpus_synced_x86_64
#define tlb_flush_page_by_mmuidx tlb_flush_page_by_mmuidx_x86_64
#define tlb_flush_page tlb_flush_page_x86_64
#define tlb_flush_page_by_mmuidx_all_cpus tlb_flush_page_by_mmuidx_all_cpus_x86_64
#define tlb_flush_page_all_cpus tlb_flush_page_all_cpus_x86_64
#define tlb_flush_page_by_mmuidx_all_cpus_synced tlb_flush_page_by_mmuidx_all_cpus_synced_x86_64
#define tlb_flush_page_all_cpus_synced tlb_flush_page_all_cpus_synced_x86_64
#define tlb_protect_code tlb_protect_code_x86_64
#define tlb_unprotect_code tlb_unprotect_code_x86_64
#define tlb_reset_dirty tlb_reset_dirty_x86_64
#define tlb_set_dirty tlb_set_dirty_x86_64
#define tlb_set_page_with_attrs tlb_set_page_with_attrs_x86_64
#define tlb_set_page tlb_set_page_x86_64
#define get_page_addr_code_hostp get_page_addr_code_hostp_x86_64
#define get_page_addr_code get_page_addr_code_x86_64
#define probe_access probe_access_x86_64
#define tlb_vaddr_to_host tlb_vaddr_to_host_x86_64
#define tlb_vaddr_to_paddr tlb_vaddr_to_paddr_x86_64
#define helper_ret_ldub_mmu helper_ret_ldub_mmu_x86_64
#define helper_le_lduw_mmu helper_le_lduw_mmu_x86_64
#define helper_be_lduw_mmu helper_be_lduw_mmu_x86_64
#define helper_le_ldul_mmu helper_le_ldul_mmu_x86_64
#define helper_be_ldul_mmu helper_be_ldul_mmu_x86_64
#define helper_le_ldq_mmu helper_le_ldq_mmu_x86_64
#define helper_be_ldq_mmu helper_be_ldq_mmu_x86_64
#define helper_ret_ldsb_mmu helper_ret_ldsb_mmu_x86_64
#define helper_le_ldsw_mmu helper_le_ldsw_mmu_x86_64
#define helper_be_ldsw_mmu helper_be_ldsw_mmu_x86_64
#define helper_le_ldsl_mmu helper_le_ldsl_mmu_x86_64
#define helper_be_ldsl_mmu helper_be_ldsl_mmu_x86_64
#define cpu_ldub_mmuidx_ra cpu_ldub_mmuidx_ra_x86_64
#define cpu_ldsb_mmuidx_ra cpu_ldsb_mmuidx_ra_x86_64
#define cpu_lduw_mmuidx_ra cpu_lduw_mmuidx_ra_x86_64
#define cpu_ldsw_mmuidx_ra cpu_ldsw_mmuidx_ra_x86_64
#define cpu_ldl_mmuidx_ra cpu_ldl_mmuidx_ra_x86_64
#define cpu_ldq_mmuidx_ra cpu_ldq_mmuidx_ra_x86_64
#define cpu_ldub_data_ra cpu_ldub_data_ra_x86_64
#define cpu_ldsb_data_ra cpu_ldsb_data_ra_x86_64
#define cpu_lduw_data_ra cpu_lduw_data_ra_x86_64
#define cpu_ldsw_data_ra cpu_ldsw_data_ra_x86_64
#define cpu_ldl_data_ra cpu_ldl_data_ra_x86_64
#define cpu_ldq_data_ra cpu_ldq_data_ra_x86_64
#define cpu_ldub_data cpu_ldub_data_x86_64
#define cpu_ldsb_data cpu_ldsb_data_x86_64
#define cpu_lduw_data cpu_lduw_data_x86_64
#define cpu_ldsw_data cpu_ldsw_data_x86_64
#define cpu_ldl_data cpu_ldl_data_x86_64
#define cpu_ldq_data cpu_ldq_data_x86_64
#define helper_ret_stb_mmu helper_ret_stb_mmu_x86_64
#define helper_le_stw_mmu helper_le_stw_mmu_x86_64
#define helper_be_stw_mmu helper_be_stw_mmu_x86_64
#define helper_le_stl_mmu helper_le_stl_mmu_x86_64
#define helper_be_stl_mmu helper_be_stl_mmu_x86_64
#define helper_le_stq_mmu helper_le_stq_mmu_x86_64
#define helper_be_stq_mmu helper_be_stq_mmu_x86_64
#define cpu_stb_mmuidx_ra cpu_stb_mmuidx_ra_x86_64
#define cpu_stw_mmuidx_ra cpu_stw_mmuidx_ra_x86_64
#define cpu_stl_mmuidx_ra cpu_stl_mmuidx_ra_x86_64
#define cpu_stq_mmuidx_ra cpu_stq_mmuidx_ra_x86_64
#define cpu_stb_data_ra cpu_stb_data_ra_x86_64
#define cpu_stw_data_ra cpu_stw_data_ra_x86_64
#define cpu_stl_data_ra cpu_stl_data_ra_x86_64
#define cpu_stq_data_ra cpu_stq_data_ra_x86_64
#define cpu_stb_data cpu_stb_data_x86_64
#define cpu_stw_data cpu_stw_data_x86_64
#define cpu_stl_data cpu_stl_data_x86_64
#define cpu_stq_data cpu_stq_data_x86_64
#define helper_atomic_cmpxchgb_mmu helper_atomic_cmpxchgb_mmu_x86_64
#define helper_atomic_xchgb_mmu helper_atomic_xchgb_mmu_x86_64
#define helper_atomic_fetch_addb_mmu helper_atomic_fetch_addb_mmu_x86_64
#define helper_atomic_fetch_andb_mmu helper_atomic_fetch_andb_mmu_x86_64
#define helper_atomic_fetch_orb_mmu helper_atomic_fetch_orb_mmu_x86_64
#define helper_atomic_fetch_xorb_mmu helper_atomic_fetch_xorb_mmu_x86_64
#define helper_atomic_add_fetchb_mmu helper_atomic_add_fetchb_mmu_x86_64
#define helper_atomic_and_fetchb_mmu helper_atomic_and_fetchb_mmu_x86_64
#define helper_atomic_or_fetchb_mmu helper_atomic_or_fetchb_mmu_x86_64
#define helper_atomic_xor_fetchb_mmu helper_atomic_xor_fetchb_mmu_x86_64
#define helper_atomic_fetch_sminb_mmu helper_atomic_fetch_sminb_mmu_x86_64
#define helper_atomic_fetch_uminb_mmu helper_atomic_fetch_uminb_mmu_x86_64
#define helper_atomic_fetch_smaxb_mmu helper_atomic_fetch_smaxb_mmu_x86_64
#define helper_atomic_fetch_umaxb_mmu helper_atomic_fetch_umaxb_mmu_x86_64
#define helper_atomic_smin_fetchb_mmu helper_atomic_smin_fetchb_mmu_x86_64
#define helper_atomic_umin_fetchb_mmu helper_atomic_umin_fetchb_mmu_x86_64
#define helper_atomic_smax_fetchb_mmu helper_atomic_smax_fetchb_mmu_x86_64
#define helper_atomic_umax_fetchb_mmu helper_atomic_umax_fetchb_mmu_x86_64
#define helper_atomic_cmpxchgw_le_mmu helper_atomic_cmpxchgw_le_mmu_x86_64
#define helper_atomic_xchgw_le_mmu helper_atomic_xchgw_le_mmu_x86_64
#define helper_atomic_fetch_addw_le_mmu helper_atomic_fetch_addw_le_mmu_x86_64
#define helper_atomic_fetch_andw_le_mmu helper_atomic_fetch_andw_le_mmu_x86_64
#define helper_atomic_fetch_orw_le_mmu helper_atomic_fetch_orw_le_mmu_x86_64
#define helper_atomic_fetch_xorw_le_mmu helper_atomic_fetch_xorw_le_mmu_x86_64
#define helper_atomic_add_fetchw_le_mmu helper_atomic_add_fetchw_le_mmu_x86_64
#define helper_atomic_and_fetchw_le_mmu helper_atomic_and_fetchw_le_mmu_x86_64
#define helper_atomic_or_fetchw_le_mmu helper_atomic_or_fetchw_le_mmu_x86_64
#define helper_atomic_xor_fetchw_le_mmu helper_atomic_xor_fetchw_le_mmu_x86_64
#define helper_atomic_fetch_sminw_le_mmu helper_atomic_fetch_sminw_le_mmu_x86_64
#define helper_atomic_fetch_uminw_le_mmu helper_atomic_fetch_uminw_le_mmu_x86_64
#define helper_atomic_fetch_smaxw_le_mmu helper_atomic_fetch_smaxw_le_mmu_x86_64
#define helper_atomic_fetch_umaxw_le_mmu helper_atomic_fetch_umaxw_le_mmu_x86_64
#define helper_atomic_smin_fetchw_le_mmu helper_atomic_smin_fetchw_le_mmu_x86_64
#define helper_atomic_umin_fetchw_le_mmu helper_atomic_umin_fetchw_le_mmu_x86_64
#define helper_atomic_smax_fetchw_le_mmu helper_atomic_smax_fetchw_le_mmu_x86_64
#define helper_atomic_umax_fetchw_le_mmu helper_atomic_umax_fetchw_le_mmu_x86_64
#define helper_atomic_cmpxchgw_be_mmu helper_atomic_cmpxchgw_be_mmu_x86_64
#define helper_atomic_xchgw_be_mmu helper_atomic_xchgw_be_mmu_x86_64
#define helper_atomic_fetch_andw_be_mmu helper_atomic_fetch_andw_be_mmu_x86_64
#define helper_atomic_fetch_orw_be_mmu helper_atomic_fetch_orw_be_mmu_x86_64
#define helper_atomic_fetch_xorw_be_mmu helper_atomic_fetch_xorw_be_mmu_x86_64
#define helper_atomic_and_fetchw_be_mmu helper_atomic_and_fetchw_be_mmu_x86_64
#define helper_atomic_or_fetchw_be_mmu helper_atomic_or_fetchw_be_mmu_x86_64
#define helper_atomic_xor_fetchw_be_mmu helper_atomic_xor_fetchw_be_mmu_x86_64
#define helper_atomic_fetch_sminw_be_mmu helper_atomic_fetch_sminw_be_mmu_x86_64
#define helper_atomic_fetch_uminw_be_mmu helper_atomic_fetch_uminw_be_mmu_x86_64
#define helper_atomic_fetch_smaxw_be_mmu helper_atomic_fetch_smaxw_be_mmu_x86_64
#define helper_atomic_fetch_umaxw_be_mmu helper_atomic_fetch_umaxw_be_mmu_x86_64
#define helper_atomic_smin_fetchw_be_mmu helper_atomic_smin_fetchw_be_mmu_x86_64
#define helper_atomic_umin_fetchw_be_mmu helper_atomic_umin_fetchw_be_mmu_x86_64
#define helper_atomic_smax_fetchw_be_mmu helper_atomic_smax_fetchw_be_mmu_x86_64
#define helper_atomic_umax_fetchw_be_mmu helper_atomic_umax_fetchw_be_mmu_x86_64
#define helper_atomic_fetch_addw_be_mmu helper_atomic_fetch_addw_be_mmu_x86_64
#define helper_atomic_add_fetchw_be_mmu helper_atomic_add_fetchw_be_mmu_x86_64
#define helper_atomic_cmpxchgl_le_mmu helper_atomic_cmpxchgl_le_mmu_x86_64
#define helper_atomic_xchgl_le_mmu helper_atomic_xchgl_le_mmu_x86_64
#define helper_atomic_fetch_addl_le_mmu helper_atomic_fetch_addl_le_mmu_x86_64
#define helper_atomic_fetch_andl_le_mmu helper_atomic_fetch_andl_le_mmu_x86_64
#define helper_atomic_fetch_orl_le_mmu helper_atomic_fetch_orl_le_mmu_x86_64
#define helper_atomic_fetch_xorl_le_mmu helper_atomic_fetch_xorl_le_mmu_x86_64
#define helper_atomic_add_fetchl_le_mmu helper_atomic_add_fetchl_le_mmu_x86_64
#define helper_atomic_and_fetchl_le_mmu helper_atomic_and_fetchl_le_mmu_x86_64
#define helper_atomic_or_fetchl_le_mmu helper_atomic_or_fetchl_le_mmu_x86_64
#define helper_atomic_xor_fetchl_le_mmu helper_atomic_xor_fetchl_le_mmu_x86_64
#define helper_atomic_fetch_sminl_le_mmu helper_atomic_fetch_sminl_le_mmu_x86_64
#define helper_atomic_fetch_uminl_le_mmu helper_atomic_fetch_uminl_le_mmu_x86_64
#define helper_atomic_fetch_smaxl_le_mmu helper_atomic_fetch_smaxl_le_mmu_x86_64
#define helper_atomic_fetch_umaxl_le_mmu helper_atomic_fetch_umaxl_le_mmu_x86_64
#define helper_atomic_smin_fetchl_le_mmu helper_atomic_smin_fetchl_le_mmu_x86_64
#define helper_atomic_umin_fetchl_le_mmu helper_atomic_umin_fetchl_le_mmu_x86_64
#define helper_atomic_smax_fetchl_le_mmu helper_atomic_smax_fetchl_le_mmu_x86_64
#define helper_atomic_umax_fetchl_le_mmu helper_atomic_umax_fetchl_le_mmu_x86_64
#define helper_atomic_cmpxchgl_be_mmu helper_atomic_cmpxchgl_be_mmu_x86_64
#define helper_atomic_xchgl_be_mmu helper_atomic_xchgl_be_mmu_x86_64
#define helper_atomic_fetch_andl_be_mmu helper_atomic_fetch_andl_be_mmu_x86_64
#define helper_atomic_fetch_orl_be_mmu helper_atomic_fetch_orl_be_mmu_x86_64
#define helper_atomic_fetch_xorl_be_mmu helper_atomic_fetch_xorl_be_mmu_x86_64
#define helper_atomic_and_fetchl_be_mmu helper_atomic_and_fetchl_be_mmu_x86_64
#define helper_atomic_or_fetchl_be_mmu helper_atomic_or_fetchl_be_mmu_x86_64
#define helper_atomic_xor_fetchl_be_mmu helper_atomic_xor_fetchl_be_mmu_x86_64
#define helper_atomic_fetch_sminl_be_mmu helper_atomic_fetch_sminl_be_mmu_x86_64
#define helper_atomic_fetch_uminl_be_mmu helper_atomic_fetch_uminl_be_mmu_x86_64
#define helper_atomic_fetch_smaxl_be_mmu helper_atomic_fetch_smaxl_be_mmu_x86_64
#define helper_atomic_fetch_umaxl_be_mmu helper_atomic_fetch_umaxl_be_mmu_x86_64
#define helper_atomic_smin_fetchl_be_mmu helper_atomic_smin_fetchl_be_mmu_x86_64
#define helper_atomic_umin_fetchl_be_mmu helper_atomic_umin_fetchl_be_mmu_x86_64
#define helper_atomic_smax_fetchl_be_mmu helper_atomic_smax_fetchl_be_mmu_x86_64
#define helper_atomic_umax_fetchl_be_mmu helper_atomic_umax_fetchl_be_mmu_x86_64
#define helper_atomic_fetch_addl_be_mmu helper_atomic_fetch_addl_be_mmu_x86_64
#define helper_atomic_add_fetchl_be_mmu helper_atomic_add_fetchl_be_mmu_x86_64
#define helper_atomic_cmpxchgq_le_mmu helper_atomic_cmpxchgq_le_mmu_x86_64
#define helper_atomic_xchgq_le_mmu helper_atomic_xchgq_le_mmu_x86_64
#define helper_atomic_fetch_addq_le_mmu helper_atomic_fetch_addq_le_mmu_x86_64
#define helper_atomic_fetch_andq_le_mmu helper_atomic_fetch_andq_le_mmu_x86_64
#define helper_atomic_fetch_orq_le_mmu helper_atomic_fetch_orq_le_mmu_x86_64
#define helper_atomic_fetch_xorq_le_mmu helper_atomic_fetch_xorq_le_mmu_x86_64
#define helper_atomic_add_fetchq_le_mmu helper_atomic_add_fetchq_le_mmu_x86_64
#define helper_atomic_and_fetchq_le_mmu helper_atomic_and_fetchq_le_mmu_x86_64
#define helper_atomic_or_fetchq_le_mmu helper_atomic_or_fetchq_le_mmu_x86_64
#define helper_atomic_xor_fetchq_le_mmu helper_atomic_xor_fetchq_le_mmu_x86_64
#define helper_atomic_fetch_sminq_le_mmu helper_atomic_fetch_sminq_le_mmu_x86_64
#define helper_atomic_fetch_uminq_le_mmu helper_atomic_fetch_uminq_le_mmu_x86_64
#define helper_atomic_fetch_smaxq_le_mmu helper_atomic_fetch_smaxq_le_mmu_x86_64
#define helper_atomic_fetch_umaxq_le_mmu helper_atomic_fetch_umaxq_le_mmu_x86_64
#define helper_atomic_smin_fetchq_le_mmu helper_atomic_smin_fetchq_le_mmu_x86_64
#define helper_atomic_umin_fetchq_le_mmu helper_atomic_umin_fetchq_le_mmu_x86_64
#define helper_atomic_smax_fetchq_le_mmu helper_atomic_smax_fetchq_le_mmu_x86_64
#define helper_atomic_umax_fetchq_le_mmu helper_atomic_umax_fetchq_le_mmu_x86_64
#define helper_atomic_cmpxchgq_be_mmu helper_atomic_cmpxchgq_be_mmu_x86_64
#define helper_atomic_xchgq_be_mmu helper_atomic_xchgq_be_mmu_x86_64
#define helper_atomic_fetch_andq_be_mmu helper_atomic_fetch_andq_be_mmu_x86_64
#define helper_atomic_fetch_orq_be_mmu helper_atomic_fetch_orq_be_mmu_x86_64
#define helper_atomic_fetch_xorq_be_mmu helper_atomic_fetch_xorq_be_mmu_x86_64
#define helper_atomic_and_fetchq_be_mmu helper_atomic_and_fetchq_be_mmu_x86_64
#define helper_atomic_or_fetchq_be_mmu helper_atomic_or_fetchq_be_mmu_x86_64
#define helper_atomic_xor_fetchq_be_mmu helper_atomic_xor_fetchq_be_mmu_x86_64
#define helper_atomic_fetch_sminq_be_mmu helper_atomic_fetch_sminq_be_mmu_x86_64
#define helper_atomic_fetch_uminq_be_mmu helper_atomic_fetch_uminq_be_mmu_x86_64
#define helper_atomic_fetch_smaxq_be_mmu helper_atomic_fetch_smaxq_be_mmu_x86_64
#define helper_atomic_fetch_umaxq_be_mmu helper_atomic_fetch_umaxq_be_mmu_x86_64
#define helper_atomic_smin_fetchq_be_mmu helper_atomic_smin_fetchq_be_mmu_x86_64
#define helper_atomic_umin_fetchq_be_mmu helper_atomic_umin_fetchq_be_mmu_x86_64
#define helper_atomic_smax_fetchq_be_mmu helper_atomic_smax_fetchq_be_mmu_x86_64
#define helper_atomic_umax_fetchq_be_mmu helper_atomic_umax_fetchq_be_mmu_x86_64
#define helper_atomic_fetch_addq_be_mmu helper_atomic_fetch_addq_be_mmu_x86_64
#define helper_atomic_add_fetchq_be_mmu helper_atomic_add_fetchq_be_mmu_x86_64
#define helper_atomic_cmpxchgb helper_atomic_cmpxchgb_x86_64
#define helper_atomic_xchgb helper_atomic_xchgb_x86_64
#define helper_atomic_fetch_addb helper_atomic_fetch_addb_x86_64
#define helper_atomic_fetch_andb helper_atomic_fetch_andb_x86_64
#define helper_atomic_fetch_orb helper_atomic_fetch_orb_x86_64
#define helper_atomic_fetch_xorb helper_atomic_fetch_xorb_x86_64
#define helper_atomic_add_fetchb helper_atomic_add_fetchb_x86_64
#define helper_atomic_and_fetchb helper_atomic_and_fetchb_x86_64
#define helper_atomic_or_fetchb helper_atomic_or_fetchb_x86_64
#define helper_atomic_xor_fetchb helper_atomic_xor_fetchb_x86_64
#define helper_atomic_fetch_sminb helper_atomic_fetch_sminb_x86_64
#define helper_atomic_fetch_uminb helper_atomic_fetch_uminb_x86_64
#define helper_atomic_fetch_smaxb helper_atomic_fetch_smaxb_x86_64
#define helper_atomic_fetch_umaxb helper_atomic_fetch_umaxb_x86_64
#define helper_atomic_smin_fetchb helper_atomic_smin_fetchb_x86_64
#define helper_atomic_umin_fetchb helper_atomic_umin_fetchb_x86_64
#define helper_atomic_smax_fetchb helper_atomic_smax_fetchb_x86_64
#define helper_atomic_umax_fetchb helper_atomic_umax_fetchb_x86_64
#define helper_atomic_cmpxchgw_le helper_atomic_cmpxchgw_le_x86_64
#define helper_atomic_xchgw_le helper_atomic_xchgw_le_x86_64
#define helper_atomic_fetch_addw_le helper_atomic_fetch_addw_le_x86_64
#define helper_atomic_fetch_andw_le helper_atomic_fetch_andw_le_x86_64
#define helper_atomic_fetch_orw_le helper_atomic_fetch_orw_le_x86_64
#define helper_atomic_fetch_xorw_le helper_atomic_fetch_xorw_le_x86_64
#define helper_atomic_add_fetchw_le helper_atomic_add_fetchw_le_x86_64
#define helper_atomic_and_fetchw_le helper_atomic_and_fetchw_le_x86_64
#define helper_atomic_or_fetchw_le helper_atomic_or_fetchw_le_x86_64
#define helper_atomic_xor_fetchw_le helper_atomic_xor_fetchw_le_x86_64
#define helper_atomic_fetch_sminw_le helper_atomic_fetch_sminw_le_x86_64
#define helper_atomic_fetch_uminw_le helper_atomic_fetch_uminw_le_x86_64
#define helper_atomic_fetch_smaxw_le helper_atomic_fetch_smaxw_le_x86_64
#define helper_atomic_fetch_umaxw_le helper_atomic_fetch_umaxw_le_x86_64
#define helper_atomic_smin_fetchw_le helper_atomic_smin_fetchw_le_x86_64
#define helper_atomic_umin_fetchw_le helper_atomic_umin_fetchw_le_x86_64
#define helper_atomic_smax_fetchw_le helper_atomic_smax_fetchw_le_x86_64
#define helper_atomic_umax_fetchw_le helper_atomic_umax_fetchw_le_x86_64
#define helper_atomic_cmpxchgw_be helper_atomic_cmpxchgw_be_x86_64
#define helper_atomic_xchgw_be helper_atomic_xchgw_be_x86_64
#define helper_atomic_fetch_andw_be helper_atomic_fetch_andw_be_x86_64
#define helper_atomic_fetch_orw_be helper_atomic_fetch_orw_be_x86_64
#define helper_atomic_fetch_xorw_be helper_atomic_fetch_xorw_be_x86_64
#define helper_atomic_and_fetchw_be helper_atomic_and_fetchw_be_x86_64
#define helper_atomic_or_fetchw_be helper_atomic_or_fetchw_be_x86_64
#define helper_atomic_xor_fetchw_be helper_atomic_xor_fetchw_be_x86_64
#define helper_atomic_fetch_sminw_be helper_atomic_fetch_sminw_be_x86_64
#define helper_atomic_fetch_uminw_be helper_atomic_fetch_uminw_be_x86_64
#define helper_atomic_fetch_smaxw_be helper_atomic_fetch_smaxw_be_x86_64
#define helper_atomic_fetch_umaxw_be helper_atomic_fetch_umaxw_be_x86_64
#define helper_atomic_smin_fetchw_be helper_atomic_smin_fetchw_be_x86_64
#define helper_atomic_umin_fetchw_be helper_atomic_umin_fetchw_be_x86_64
#define helper_atomic_smax_fetchw_be helper_atomic_smax_fetchw_be_x86_64
#define helper_atomic_umax_fetchw_be helper_atomic_umax_fetchw_be_x86_64
#define helper_atomic_fetch_addw_be helper_atomic_fetch_addw_be_x86_64
#define helper_atomic_add_fetchw_be helper_atomic_add_fetchw_be_x86_64
#define helper_atomic_cmpxchgl_le helper_atomic_cmpxchgl_le_x86_64
#define helper_atomic_xchgl_le helper_atomic_xchgl_le_x86_64
#define helper_atomic_fetch_addl_le helper_atomic_fetch_addl_le_x86_64
#define helper_atomic_fetch_andl_le helper_atomic_fetch_andl_le_x86_64
#define helper_atomic_fetch_orl_le helper_atomic_fetch_orl_le_x86_64
#define helper_atomic_fetch_xorl_le helper_atomic_fetch_xorl_le_x86_64
#define helper_atomic_add_fetchl_le helper_atomic_add_fetchl_le_x86_64
#define helper_atomic_and_fetchl_le helper_atomic_and_fetchl_le_x86_64
#define helper_atomic_or_fetchl_le helper_atomic_or_fetchl_le_x86_64
#define helper_atomic_xor_fetchl_le helper_atomic_xor_fetchl_le_x86_64
#define helper_atomic_fetch_sminl_le helper_atomic_fetch_sminl_le_x86_64
#define helper_atomic_fetch_uminl_le helper_atomic_fetch_uminl_le_x86_64
#define helper_atomic_fetch_smaxl_le helper_atomic_fetch_smaxl_le_x86_64
#define helper_atomic_fetch_umaxl_le helper_atomic_fetch_umaxl_le_x86_64
#define helper_atomic_smin_fetchl_le helper_atomic_smin_fetchl_le_x86_64
#define helper_atomic_umin_fetchl_le helper_atomic_umin_fetchl_le_x86_64
#define helper_atomic_smax_fetchl_le helper_atomic_smax_fetchl_le_x86_64
#define helper_atomic_umax_fetchl_le helper_atomic_umax_fetchl_le_x86_64
#define helper_atomic_cmpxchgl_be helper_atomic_cmpxchgl_be_x86_64
#define helper_atomic_xchgl_be helper_atomic_xchgl_be_x86_64
#define helper_atomic_fetch_andl_be helper_atomic_fetch_andl_be_x86_64
#define helper_atomic_fetch_orl_be helper_atomic_fetch_orl_be_x86_64
#define helper_atomic_fetch_xorl_be helper_atomic_fetch_xorl_be_x86_64
#define helper_atomic_and_fetchl_be helper_atomic_and_fetchl_be_x86_64
#define helper_atomic_or_fetchl_be helper_atomic_or_fetchl_be_x86_64
#define helper_atomic_xor_fetchl_be helper_atomic_xor_fetchl_be_x86_64
#define helper_atomic_fetch_sminl_be helper_atomic_fetch_sminl_be_x86_64
#define helper_atomic_fetch_uminl_be helper_atomic_fetch_uminl_be_x86_64
#define helper_atomic_fetch_smaxl_be helper_atomic_fetch_smaxl_be_x86_64
#define helper_atomic_fetch_umaxl_be helper_atomic_fetch_umaxl_be_x86_64
#define helper_atomic_smin_fetchl_be helper_atomic_smin_fetchl_be_x86_64
#define helper_atomic_umin_fetchl_be helper_atomic_umin_fetchl_be_x86_64
#define helper_atomic_smax_fetchl_be helper_atomic_smax_fetchl_be_x86_64
#define helper_atomic_umax_fetchl_be helper_atomic_umax_fetchl_be_x86_64
#define helper_atomic_fetch_addl_be helper_atomic_fetch_addl_be_x86_64
#define helper_atomic_add_fetchl_be helper_atomic_add_fetchl_be_x86_64
#define helper_atomic_cmpxchgq_le helper_atomic_cmpxchgq_le_x86_64
#define helper_atomic_xchgq_le helper_atomic_xchgq_le_x86_64
#define helper_atomic_fetch_addq_le helper_atomic_fetch_addq_le_x86_64
#define helper_atomic_fetch_andq_le helper_atomic_fetch_andq_le_x86_64
#define helper_atomic_fetch_orq_le helper_atomic_fetch_orq_le_x86_64
#define helper_atomic_fetch_xorq_le helper_atomic_fetch_xorq_le_x86_64
#define helper_atomic_add_fetchq_le helper_atomic_add_fetchq_le_x86_64
#define helper_atomic_and_fetchq_le helper_atomic_and_fetchq_le_x86_64
#define helper_atomic_or_fetchq_le helper_atomic_or_fetchq_le_x86_64
#define helper_atomic_xor_fetchq_le helper_atomic_xor_fetchq_le_x86_64
#define helper_atomic_fetch_sminq_le helper_atomic_fetch_sminq_le_x86_64
#define helper_atomic_fetch_uminq_le helper_atomic_fetch_uminq_le_x86_64
#define helper_atomic_fetch_smaxq_le helper_atomic_fetch_smaxq_le_x86_64
#define helper_atomic_fetch_umaxq_le helper_atomic_fetch_umaxq_le_x86_64
#define helper_atomic_smin_fetchq_le helper_atomic_smin_fetchq_le_x86_64
#define helper_atomic_umin_fetchq_le helper_atomic_umin_fetchq_le_x86_64
#define helper_atomic_smax_fetchq_le helper_atomic_smax_fetchq_le_x86_64
#define helper_atomic_umax_fetchq_le helper_atomic_umax_fetchq_le_x86_64
#define helper_atomic_cmpxchgq_be helper_atomic_cmpxchgq_be_x86_64
#define helper_atomic_xchgq_be helper_atomic_xchgq_be_x86_64
#define helper_atomic_fetch_andq_be helper_atomic_fetch_andq_be_x86_64
#define helper_atomic_fetch_orq_be helper_atomic_fetch_orq_be_x86_64
#define helper_atomic_fetch_xorq_be helper_atomic_fetch_xorq_be_x86_64
#define helper_atomic_and_fetchq_be helper_atomic_and_fetchq_be_x86_64
#define helper_atomic_or_fetchq_be helper_atomic_or_fetchq_be_x86_64
#define helper_atomic_xor_fetchq_be helper_atomic_xor_fetchq_be_x86_64
#define helper_atomic_fetch_sminq_be helper_atomic_fetch_sminq_be_x86_64
#define helper_atomic_fetch_uminq_be helper_atomic_fetch_uminq_be_x86_64
#define helper_atomic_fetch_smaxq_be helper_atomic_fetch_smaxq_be_x86_64
#define helper_atomic_fetch_umaxq_be helper_atomic_fetch_umaxq_be_x86_64
#define helper_atomic_smin_fetchq_be helper_atomic_smin_fetchq_be_x86_64
#define helper_atomic_umin_fetchq_be helper_atomic_umin_fetchq_be_x86_64
#define helper_atomic_smax_fetchq_be helper_atomic_smax_fetchq_be_x86_64
#define helper_atomic_umax_fetchq_be helper_atomic_umax_fetchq_be_x86_64
#define helper_atomic_fetch_addq_be helper_atomic_fetch_addq_be_x86_64
#define helper_atomic_add_fetchq_be helper_atomic_add_fetchq_be_x86_64
#define cpu_ldub_code cpu_ldub_code_x86_64
#define cpu_lduw_code cpu_lduw_code_x86_64
#define cpu_ldl_code cpu_ldl_code_x86_64
#define cpu_ldq_code cpu_ldq_code_x86_64
#define helper_div_i32 helper_div_i32_x86_64
#define helper_rem_i32 helper_rem_i32_x86_64
#define helper_divu_i32 helper_divu_i32_x86_64
#define helper_remu_i32 helper_remu_i32_x86_64
#define helper_shl_i64 helper_shl_i64_x86_64
#define helper_shr_i64 helper_shr_i64_x86_64
#define helper_sar_i64 helper_sar_i64_x86_64
#define helper_div_i64 helper_div_i64_x86_64
#define helper_rem_i64 helper_rem_i64_x86_64
#define helper_divu_i64 helper_divu_i64_x86_64
#define helper_remu_i64 helper_remu_i64_x86_64
#define helper_muluh_i64 helper_muluh_i64_x86_64
#define helper_mulsh_i64 helper_mulsh_i64_x86_64
#define helper_clz_i32 helper_clz_i32_x86_64
#define helper_ctz_i32 helper_ctz_i32_x86_64
#define helper_clz_i64 helper_clz_i64_x86_64
#define helper_ctz_i64 helper_ctz_i64_x86_64
#define helper_clrsb_i32 helper_clrsb_i32_x86_64
#define helper_clrsb_i64 helper_clrsb_i64_x86_64
#define helper_ctpop_i32 helper_ctpop_i32_x86_64
#define helper_ctpop_i64 helper_ctpop_i64_x86_64
#define helper_lookup_tb_ptr helper_lookup_tb_ptr_x86_64
#define helper_exit_atomic helper_exit_atomic_x86_64
#define helper_gvec_add8 helper_gvec_add8_x86_64
#define helper_gvec_add16 helper_gvec_add16_x86_64
#define helper_gvec_add32 helper_gvec_add32_x86_64
#define helper_gvec_add64 helper_gvec_add64_x86_64
#define helper_gvec_adds8 helper_gvec_adds8_x86_64
#define helper_gvec_adds16 helper_gvec_adds16_x86_64
#define helper_gvec_adds32 helper_gvec_adds32_x86_64
#define helper_gvec_adds64 helper_gvec_adds64_x86_64
#define helper_gvec_sub8 helper_gvec_sub8_x86_64
#define helper_gvec_sub16 helper_gvec_sub16_x86_64
#define helper_gvec_sub32 helper_gvec_sub32_x86_64
#define helper_gvec_sub64 helper_gvec_sub64_x86_64
#define helper_gvec_subs8 helper_gvec_subs8_x86_64
#define helper_gvec_subs16 helper_gvec_subs16_x86_64
#define helper_gvec_subs32 helper_gvec_subs32_x86_64
#define helper_gvec_subs64 helper_gvec_subs64_x86_64
#define helper_gvec_mul8 helper_gvec_mul8_x86_64
#define helper_gvec_mul16 helper_gvec_mul16_x86_64
#define helper_gvec_mul32 helper_gvec_mul32_x86_64
#define helper_gvec_mul64 helper_gvec_mul64_x86_64
#define helper_gvec_muls8 helper_gvec_muls8_x86_64
#define helper_gvec_muls16 helper_gvec_muls16_x86_64
#define helper_gvec_muls32 helper_gvec_muls32_x86_64
#define helper_gvec_muls64 helper_gvec_muls64_x86_64
#define helper_gvec_neg8 helper_gvec_neg8_x86_64
#define helper_gvec_neg16 helper_gvec_neg16_x86_64
#define helper_gvec_neg32 helper_gvec_neg32_x86_64
#define helper_gvec_neg64 helper_gvec_neg64_x86_64
#define helper_gvec_abs8 helper_gvec_abs8_x86_64
#define helper_gvec_abs16 helper_gvec_abs16_x86_64
#define helper_gvec_abs32 helper_gvec_abs32_x86_64
#define helper_gvec_abs64 helper_gvec_abs64_x86_64
#define helper_gvec_mov helper_gvec_mov_x86_64
#define helper_gvec_dup64 helper_gvec_dup64_x86_64
#define helper_gvec_dup32 helper_gvec_dup32_x86_64
#define helper_gvec_dup16 helper_gvec_dup16_x86_64
#define helper_gvec_dup8 helper_gvec_dup8_x86_64
#define helper_gvec_not helper_gvec_not_x86_64
#define helper_gvec_and helper_gvec_and_x86_64
#define helper_gvec_or helper_gvec_or_x86_64
#define helper_gvec_xor helper_gvec_xor_x86_64
#define helper_gvec_andc helper_gvec_andc_x86_64
#define helper_gvec_orc helper_gvec_orc_x86_64
#define helper_gvec_nand helper_gvec_nand_x86_64
#define helper_gvec_nor helper_gvec_nor_x86_64
#define helper_gvec_eqv helper_gvec_eqv_x86_64
#define helper_gvec_ands helper_gvec_ands_x86_64
#define helper_gvec_xors helper_gvec_xors_x86_64
#define helper_gvec_ors helper_gvec_ors_x86_64
#define helper_gvec_shl8i helper_gvec_shl8i_x86_64
#define helper_gvec_shl16i helper_gvec_shl16i_x86_64
#define helper_gvec_shl32i helper_gvec_shl32i_x86_64
#define helper_gvec_shl64i helper_gvec_shl64i_x86_64
#define helper_gvec_shr8i helper_gvec_shr8i_x86_64
#define helper_gvec_shr16i helper_gvec_shr16i_x86_64
#define helper_gvec_shr32i helper_gvec_shr32i_x86_64
#define helper_gvec_shr64i helper_gvec_shr64i_x86_64
#define helper_gvec_sar8i helper_gvec_sar8i_x86_64
#define helper_gvec_sar16i helper_gvec_sar16i_x86_64
#define helper_gvec_sar32i helper_gvec_sar32i_x86_64
#define helper_gvec_sar64i helper_gvec_sar64i_x86_64
#define helper_gvec_shl8v helper_gvec_shl8v_x86_64
#define helper_gvec_shl16v helper_gvec_shl16v_x86_64
#define helper_gvec_shl32v helper_gvec_shl32v_x86_64
#define helper_gvec_shl64v helper_gvec_shl64v_x86_64
#define helper_gvec_shr8v helper_gvec_shr8v_x86_64
#define helper_gvec_shr16v helper_gvec_shr16v_x86_64
#define helper_gvec_shr32v helper_gvec_shr32v_x86_64
#define helper_gvec_shr64v helper_gvec_shr64v_x86_64
#define helper_gvec_sar8v helper_gvec_sar8v_x86_64
#define helper_gvec_sar16v helper_gvec_sar16v_x86_64
#define helper_gvec_sar32v helper_gvec_sar32v_x86_64
#define helper_gvec_sar64v helper_gvec_sar64v_x86_64
#define helper_gvec_eq8 helper_gvec_eq8_x86_64
#define helper_gvec_ne8 helper_gvec_ne8_x86_64
#define helper_gvec_lt8 helper_gvec_lt8_x86_64
#define helper_gvec_le8 helper_gvec_le8_x86_64
#define helper_gvec_ltu8 helper_gvec_ltu8_x86_64
#define helper_gvec_leu8 helper_gvec_leu8_x86_64
#define helper_gvec_eq16 helper_gvec_eq16_x86_64
#define helper_gvec_ne16 helper_gvec_ne16_x86_64
#define helper_gvec_lt16 helper_gvec_lt16_x86_64
#define helper_gvec_le16 helper_gvec_le16_x86_64
#define helper_gvec_ltu16 helper_gvec_ltu16_x86_64
#define helper_gvec_leu16 helper_gvec_leu16_x86_64
#define helper_gvec_eq32 helper_gvec_eq32_x86_64
#define helper_gvec_ne32 helper_gvec_ne32_x86_64
#define helper_gvec_lt32 helper_gvec_lt32_x86_64
#define helper_gvec_le32 helper_gvec_le32_x86_64
#define helper_gvec_ltu32 helper_gvec_ltu32_x86_64
#define helper_gvec_leu32 helper_gvec_leu32_x86_64
#define helper_gvec_eq64 helper_gvec_eq64_x86_64
#define helper_gvec_ne64 helper_gvec_ne64_x86_64
#define helper_gvec_lt64 helper_gvec_lt64_x86_64
#define helper_gvec_le64 helper_gvec_le64_x86_64
#define helper_gvec_ltu64 helper_gvec_ltu64_x86_64
#define helper_gvec_leu64 helper_gvec_leu64_x86_64
#define helper_gvec_ssadd8 helper_gvec_ssadd8_x86_64
#define helper_gvec_ssadd16 helper_gvec_ssadd16_x86_64
#define helper_gvec_ssadd32 helper_gvec_ssadd32_x86_64
#define helper_gvec_ssadd64 helper_gvec_ssadd64_x86_64
#define helper_gvec_sssub8 helper_gvec_sssub8_x86_64
#define helper_gvec_sssub16 helper_gvec_sssub16_x86_64
#define helper_gvec_sssub32 helper_gvec_sssub32_x86_64
#define helper_gvec_sssub64 helper_gvec_sssub64_x86_64
#define helper_gvec_usadd8 helper_gvec_usadd8_x86_64
#define helper_gvec_usadd16 helper_gvec_usadd16_x86_64
#define helper_gvec_usadd32 helper_gvec_usadd32_x86_64
#define helper_gvec_usadd64 helper_gvec_usadd64_x86_64
#define helper_gvec_ussub8 helper_gvec_ussub8_x86_64
#define helper_gvec_ussub16 helper_gvec_ussub16_x86_64
#define helper_gvec_ussub32 helper_gvec_ussub32_x86_64
#define helper_gvec_ussub64 helper_gvec_ussub64_x86_64
#define helper_gvec_smin8 helper_gvec_smin8_x86_64
#define helper_gvec_smin16 helper_gvec_smin16_x86_64
#define helper_gvec_smin32 helper_gvec_smin32_x86_64
#define helper_gvec_smin64 helper_gvec_smin64_x86_64
#define helper_gvec_smax8 helper_gvec_smax8_x86_64
#define helper_gvec_smax16 helper_gvec_smax16_x86_64
#define helper_gvec_smax32 helper_gvec_smax32_x86_64
#define helper_gvec_smax64 helper_gvec_smax64_x86_64
#define helper_gvec_umin8 helper_gvec_umin8_x86_64
#define helper_gvec_umin16 helper_gvec_umin16_x86_64
#define helper_gvec_umin32 helper_gvec_umin32_x86_64
#define helper_gvec_umin64 helper_gvec_umin64_x86_64
#define helper_gvec_umax8 helper_gvec_umax8_x86_64
#define helper_gvec_umax16 helper_gvec_umax16_x86_64
#define helper_gvec_umax32 helper_gvec_umax32_x86_64
#define helper_gvec_umax64 helper_gvec_umax64_x86_64
#define helper_gvec_bitsel helper_gvec_bitsel_x86_64
#define cpu_restore_state cpu_restore_state_x86_64
#define page_collection_lock page_collection_lock_x86_64
#define page_collection_unlock page_collection_unlock_x86_64
#define free_code_gen_buffer free_code_gen_buffer_x86_64
#define tcg_exec_init tcg_exec_init_x86_64
#define tb_cleanup tb_cleanup_x86_64
#define tb_flush tb_flush_x86_64
#define tb_phys_invalidate tb_phys_invalidate_x86_64
#define tb_gen_code tb_gen_code_x86_64
#define tb_exec_lock tb_exec_lock_x86_64
#define tb_exec_unlock tb_exec_unlock_x86_64
#define tb_invalidate_phys_page_range tb_invalidate_phys_page_range_x86_64
#define tb_invalidate_phys_range tb_invalidate_phys_range_x86_64
#define tb_invalidate_phys_page_fast tb_invalidate_phys_page_fast_x86_64
#define tb_check_watchpoint tb_check_watchpoint_x86_64
#define cpu_io_recompile cpu_io_recompile_x86_64
#define tb_flush_jmp_cache tb_flush_jmp_cache_x86_64
#define tcg_flush_softmmu_tlb tcg_flush_softmmu_tlb_x86_64
#define translator_loop_temp_check translator_loop_temp_check_x86_64
#define translator_loop translator_loop_x86_64
#define helper_atomic_cmpxchgo_le_mmu helper_atomic_cmpxchgo_le_mmu_x86_64
#define helper_atomic_cmpxchgo_be_mmu helper_atomic_cmpxchgo_be_mmu_x86_64
#define helper_atomic_ldo_le_mmu helper_atomic_ldo_le_mmu_x86_64
#define helper_atomic_ldo_be_mmu helper_atomic_ldo_be_mmu_x86_64
#define helper_atomic_sto_le_mmu helper_atomic_sto_le_mmu_x86_64
#define helper_atomic_sto_be_mmu helper_atomic_sto_be_mmu_x86_64
#define unassigned_mem_ops unassigned_mem_ops_x86_64
#define floatx80_infinity floatx80_infinity_x86_64
#define dup_const_func dup_const_func_x86_64
#define gen_helper_raise_exception gen_helper_raise_exception_x86_64
#define gen_helper_raise_interrupt gen_helper_raise_interrupt_x86_64
#define gen_helper_vfp_get_fpscr gen_helper_vfp_get_fpscr_x86_64
#define gen_helper_vfp_set_fpscr gen_helper_vfp_set_fpscr_x86_64
#define gen_helper_cpsr_read gen_helper_cpsr_read_x86_64
#define gen_helper_cpsr_write gen_helper_cpsr_write_x86_64
#define tlb_reset_dirty_by_vaddr tlb_reset_dirty_by_vaddr_x86_64
#define helper_stqcx_le_parallel helper_stqcx_le_parallel_x86_64
#define helper_stqcx_be_parallel helper_stqcx_be_parallel_x86_64
#define cpu_get_tsc cpu_get_tsc_x86_64
#define x86_cpu_get_memory_mapping x86_cpu_get_memory_mapping_x86_64
#define cpu_x86_update_dr7 cpu_x86_update_dr7_x86_64
#define breakpoint_handler breakpoint_handler_x86_64
#define helper_single_step helper_single_step_x86_64
#define helper_rechecking_single_step helper_rechecking_single_step_x86_64
#define helper_set_dr helper_set_dr_x86_64
#define helper_get_dr helper_get_dr_x86_64
#define helper_bpt_io helper_bpt_io_x86_64
#define helper_cc_compute_all helper_cc_compute_all_x86_64
#define cpu_cc_compute_all cpu_cc_compute_all_x86_64
#define helper_cc_compute_c helper_cc_compute_c_x86_64
#define helper_write_eflags helper_write_eflags_x86_64
#define helper_read_eflags helper_read_eflags_x86_64
#define helper_clts helper_clts_x86_64
#define helper_reset_rf helper_reset_rf_x86_64
#define helper_cli helper_cli_x86_64
#define helper_sti helper_sti_x86_64
#define helper_clac helper_clac_x86_64
#define helper_stac helper_stac_x86_64
#define get_register_name_32 get_register_name_32_x86_64
#define host_cpuid host_cpuid_x86_64
#define host_vendor_fms host_vendor_fms_x86_64
#define x86_cpu_set_default_version x86_cpu_set_default_version_x86_64
#define cpu_clear_apic_feature cpu_clear_apic_feature_x86_64
#define cpu_x86_cpuid cpu_x86_cpuid_x86_64
#define x86_cpu_pending_interrupt x86_cpu_pending_interrupt_x86_64
#define x86_update_hflags x86_update_hflags_x86_64
#define cpu_x86_init cpu_x86_init_x86_64
#define helper_raise_interrupt helper_raise_interrupt_x86_64
#define helper_raise_exception helper_raise_exception_x86_64
#define raise_interrupt raise_interrupt_x86_64
#define raise_exception_err raise_exception_err_x86_64
#define raise_exception_err_ra raise_exception_err_ra_x86_64
#define raise_exception raise_exception_x86_64
#define raise_exception_ra raise_exception_ra_x86_64
#define x86_cpu_tlb_fill x86_cpu_tlb_fill_x86_64
#define cpu_set_ignne cpu_set_ignne_x86_64
#define helper_flds_FT0 helper_flds_FT0_x86_64
#define helper_fldl_FT0 helper_fldl_FT0_x86_64
#define helper_fildl_FT0 helper_fildl_FT0_x86_64
#define helper_flds_ST0 helper_flds_ST0_x86_64
#define helper_fldl_ST0 helper_fldl_ST0_x86_64
#define helper_fildl_ST0 helper_fildl_ST0_x86_64
#define helper_fildll_ST0 helper_fildll_ST0_x86_64
#define helper_fsts_ST0 helper_fsts_ST0_x86_64
#define helper_fstl_ST0 helper_fstl_ST0_x86_64
#define helper_fist_ST0 helper_fist_ST0_x86_64
#define helper_fistl_ST0 helper_fistl_ST0_x86_64
#define helper_fistll_ST0 helper_fistll_ST0_x86_64
#define helper_fistt_ST0 helper_fistt_ST0_x86_64
#define helper_fisttl_ST0 helper_fisttl_ST0_x86_64
#define helper_fisttll_ST0 helper_fisttll_ST0_x86_64
#define helper_fldt_ST0 helper_fldt_ST0_x86_64
#define helper_fstt_ST0 helper_fstt_ST0_x86_64
#define helper_fpush helper_fpush_x86_64
#define helper_fpop helper_fpop_x86_64
#define helper_fdecstp helper_fdecstp_x86_64
#define helper_fincstp helper_fincstp_x86_64
#define helper_ffree_STN helper_ffree_STN_x86_64
#define helper_fmov_ST0_FT0 helper_fmov_ST0_FT0_x86_64
#define helper_fmov_FT0_STN helper_fmov_FT0_STN_x86_64
#define helper_fmov_ST0_STN helper_fmov_ST0_STN_x86_64
#define helper_fmov_STN_ST0 helper_fmov_STN_ST0_x86_64
#define helper_fxchg_ST0_STN helper_fxchg_ST0_STN_x86_64
#define helper_fcom_ST0_FT0 helper_fcom_ST0_FT0_x86_64
#define helper_fucom_ST0_FT0 helper_fucom_ST0_FT0_x86_64
#define helper_fcomi_ST0_FT0 helper_fcomi_ST0_FT0_x86_64
#define helper_fucomi_ST0_FT0 helper_fucomi_ST0_FT0_x86_64
#define helper_fadd_ST0_FT0 helper_fadd_ST0_FT0_x86_64
#define helper_fmul_ST0_FT0 helper_fmul_ST0_FT0_x86_64
#define helper_fsub_ST0_FT0 helper_fsub_ST0_FT0_x86_64
#define helper_fsubr_ST0_FT0 helper_fsubr_ST0_FT0_x86_64
#define helper_fdiv_ST0_FT0 helper_fdiv_ST0_FT0_x86_64
#define helper_fdivr_ST0_FT0 helper_fdivr_ST0_FT0_x86_64
#define helper_fadd_STN_ST0 helper_fadd_STN_ST0_x86_64
#define helper_fmul_STN_ST0 helper_fmul_STN_ST0_x86_64
#define helper_fsub_STN_ST0 helper_fsub_STN_ST0_x86_64
#define helper_fsubr_STN_ST0 helper_fsubr_STN_ST0_x86_64
#define helper_fdiv_STN_ST0 helper_fdiv_STN_ST0_x86_64
#define helper_fdivr_STN_ST0 helper_fdivr_STN_ST0_x86_64
#define helper_fchs_ST0 helper_fchs_ST0_x86_64
#define helper_fabs_ST0 helper_fabs_ST0_x86_64
#define helper_fld1_ST0 helper_fld1_ST0_x86_64
#define helper_fldl2t_ST0 helper_fldl2t_ST0_x86_64
#define helper_fldl2e_ST0 helper_fldl2e_ST0_x86_64
#define helper_fldpi_ST0 helper_fldpi_ST0_x86_64
#define helper_fldlg2_ST0 helper_fldlg2_ST0_x86_64
#define helper_fldln2_ST0 helper_fldln2_ST0_x86_64
#define helper_fldz_ST0 helper_fldz_ST0_x86_64
#define helper_fldz_FT0 helper_fldz_FT0_x86_64
#define helper_fnstsw helper_fnstsw_x86_64
#define helper_fnstcw helper_fnstcw_x86_64
#define update_fp_status update_fp_status_x86_64
#define helper_fldcw helper_fldcw_x86_64
#define helper_fclex helper_fclex_x86_64
#define helper_fwait helper_fwait_x86_64
#define helper_fninit helper_fninit_x86_64
#define helper_fbld_ST0 helper_fbld_ST0_x86_64
#define helper_fbst_ST0 helper_fbst_ST0_x86_64
#define helper_f2xm1 helper_f2xm1_x86_64
#define helper_fyl2x helper_fyl2x_x86_64
#define helper_fptan helper_fptan_x86_64
#define helper_fpatan helper_fpatan_x86_64
#define helper_fxtract helper_fxtract_x86_64
#define helper_fprem1 helper_fprem1_x86_64
#define helper_fprem helper_fprem_x86_64
#define helper_fyl2xp1 helper_fyl2xp1_x86_64
#define helper_fsqrt helper_fsqrt_x86_64
#define helper_fsincos helper_fsincos_x86_64
#define helper_frndint helper_frndint_x86_64
#define helper_fscale helper_fscale_x86_64
#define helper_fsin helper_fsin_x86_64
#define helper_fcos helper_fcos_x86_64
#define helper_fxam_ST0 helper_fxam_ST0_x86_64
#define helper_fstenv helper_fstenv_x86_64
#define helper_fldenv helper_fldenv_x86_64
#define helper_fsave helper_fsave_x86_64
#define helper_frstor helper_frstor_x86_64
#define helper_fxsave helper_fxsave_x86_64
#define helper_xsave helper_xsave_x86_64
#define helper_xsaveopt helper_xsaveopt_x86_64
#define helper_fxrstor helper_fxrstor_x86_64
#define helper_xrstor helper_xrstor_x86_64
#define helper_xgetbv helper_xgetbv_x86_64
#define helper_xsetbv helper_xsetbv_x86_64
#define update_mxcsr_status update_mxcsr_status_x86_64
#define helper_ldmxcsr helper_ldmxcsr_x86_64
#define helper_enter_mmx helper_enter_mmx_x86_64
#define helper_emms helper_emms_x86_64
#define helper_movq helper_movq_x86_64
#define helper_psrlw_mmx helper_psrlw_mmx_x86_64
#define helper_psraw_mmx helper_psraw_mmx_x86_64
#define helper_psllw_mmx helper_psllw_mmx_x86_64
#define helper_psrld_mmx helper_psrld_mmx_x86_64
#define helper_psrad_mmx helper_psrad_mmx_x86_64
#define helper_pslld_mmx helper_pslld_mmx_x86_64
#define helper_psrlq_mmx helper_psrlq_mmx_x86_64
#define helper_psllq_mmx helper_psllq_mmx_x86_64
#define helper_paddb_mmx helper_paddb_mmx_x86_64
#define helper_paddw_mmx helper_paddw_mmx_x86_64
#define helper_paddl_mmx helper_paddl_mmx_x86_64
#define helper_paddq_mmx helper_paddq_mmx_x86_64
#define helper_psubb_mmx helper_psubb_mmx_x86_64
#define helper_psubw_mmx helper_psubw_mmx_x86_64
#define helper_psubl_mmx helper_psubl_mmx_x86_64
#define helper_psubq_mmx helper_psubq_mmx_x86_64
#define helper_paddusb_mmx helper_paddusb_mmx_x86_64
#define helper_paddsb_mmx helper_paddsb_mmx_x86_64
#define helper_psubusb_mmx helper_psubusb_mmx_x86_64
#define helper_psubsb_mmx helper_psubsb_mmx_x86_64
#define helper_paddusw_mmx helper_paddusw_mmx_x86_64
#define helper_paddsw_mmx helper_paddsw_mmx_x86_64
#define helper_psubusw_mmx helper_psubusw_mmx_x86_64
#define helper_psubsw_mmx helper_psubsw_mmx_x86_64
#define helper_pminub_mmx helper_pminub_mmx_x86_64
#define helper_pmaxub_mmx helper_pmaxub_mmx_x86_64
#define helper_pminsw_mmx helper_pminsw_mmx_x86_64
#define helper_pmaxsw_mmx helper_pmaxsw_mmx_x86_64
#define helper_pand_mmx helper_pand_mmx_x86_64
#define helper_pandn_mmx helper_pandn_mmx_x86_64
#define helper_por_mmx helper_por_mmx_x86_64
#define helper_pxor_mmx helper_pxor_mmx_x86_64
#define helper_pcmpgtb_mmx helper_pcmpgtb_mmx_x86_64
#define helper_pcmpgtw_mmx helper_pcmpgtw_mmx_x86_64
#define helper_pcmpgtl_mmx helper_pcmpgtl_mmx_x86_64
#define helper_pcmpeqb_mmx helper_pcmpeqb_mmx_x86_64
#define helper_pcmpeqw_mmx helper_pcmpeqw_mmx_x86_64
#define helper_pcmpeql_mmx helper_pcmpeql_mmx_x86_64
#define helper_pmullw_mmx helper_pmullw_mmx_x86_64
#define helper_pmulhrw_mmx helper_pmulhrw_mmx_x86_64
#define helper_pmulhuw_mmx helper_pmulhuw_mmx_x86_64
#define helper_pmulhw_mmx helper_pmulhw_mmx_x86_64
#define helper_pavgb_mmx helper_pavgb_mmx_x86_64
#define helper_pavgw_mmx helper_pavgw_mmx_x86_64
#define helper_pmuludq_mmx helper_pmuludq_mmx_x86_64
#define helper_pmaddwd_mmx helper_pmaddwd_mmx_x86_64
#define helper_psadbw_mmx helper_psadbw_mmx_x86_64
#define helper_maskmov_mmx helper_maskmov_mmx_x86_64
#define helper_movl_mm_T0_mmx helper_movl_mm_T0_mmx_x86_64
#define helper_movq_mm_T0_mmx helper_movq_mm_T0_mmx_x86_64
#define helper_pshufw_mmx helper_pshufw_mmx_x86_64
#define helper_pmovmskb_mmx helper_pmovmskb_mmx_x86_64
#define helper_packsswb_mmx helper_packsswb_mmx_x86_64
#define helper_packuswb_mmx helper_packuswb_mmx_x86_64
#define helper_packssdw_mmx helper_packssdw_mmx_x86_64
#define helper_punpcklbw_mmx helper_punpcklbw_mmx_x86_64
#define helper_punpcklwd_mmx helper_punpcklwd_mmx_x86_64
#define helper_punpckldq_mmx helper_punpckldq_mmx_x86_64
#define helper_punpckhbw_mmx helper_punpckhbw_mmx_x86_64
#define helper_punpckhwd_mmx helper_punpckhwd_mmx_x86_64
#define helper_punpckhdq_mmx helper_punpckhdq_mmx_x86_64
#define helper_pi2fd helper_pi2fd_x86_64
#define helper_pi2fw helper_pi2fw_x86_64
#define helper_pf2id helper_pf2id_x86_64
#define helper_pf2iw helper_pf2iw_x86_64
#define helper_pfacc helper_pfacc_x86_64
#define helper_pfadd helper_pfadd_x86_64
#define helper_pfcmpeq helper_pfcmpeq_x86_64
#define helper_pfcmpge helper_pfcmpge_x86_64
#define helper_pfcmpgt helper_pfcmpgt_x86_64
#define helper_pfmax helper_pfmax_x86_64
#define helper_pfmin helper_pfmin_x86_64
#define helper_pfmul helper_pfmul_x86_64
#define helper_pfnacc helper_pfnacc_x86_64
#define helper_pfpnacc helper_pfpnacc_x86_64
#define helper_pfrcp helper_pfrcp_x86_64
#define helper_pfrsqrt helper_pfrsqrt_x86_64
#define helper_pfsub helper_pfsub_x86_64
#define helper_pfsubr helper_pfsubr_x86_64
#define helper_pswapd helper_pswapd_x86_64
#define helper_pshufb_mmx helper_pshufb_mmx_x86_64
#define helper_phaddw_mmx helper_phaddw_mmx_x86_64
#define helper_phaddd_mmx helper_phaddd_mmx_x86_64
#define helper_phaddsw_mmx helper_phaddsw_mmx_x86_64
#define helper_pmaddubsw_mmx helper_pmaddubsw_mmx_x86_64
#define helper_phsubw_mmx helper_phsubw_mmx_x86_64
#define helper_phsubd_mmx helper_phsubd_mmx_x86_64
#define helper_phsubsw_mmx helper_phsubsw_mmx_x86_64
#define helper_pabsb_mmx helper_pabsb_mmx_x86_64
#define helper_pabsw_mmx helper_pabsw_mmx_x86_64
#define helper_pabsd_mmx helper_pabsd_mmx_x86_64
#define helper_pmulhrsw_mmx helper_pmulhrsw_mmx_x86_64
#define helper_psignb_mmx helper_psignb_mmx_x86_64
#define helper_psignw_mmx helper_psignw_mmx_x86_64
#define helper_psignd_mmx helper_psignd_mmx_x86_64
#define helper_palignr_mmx helper_palignr_mmx_x86_64
#define helper_psrlw_xmm helper_psrlw_xmm_x86_64
#define helper_psraw_xmm helper_psraw_xmm_x86_64
#define helper_psllw_xmm helper_psllw_xmm_x86_64
#define helper_psrld_xmm helper_psrld_xmm_x86_64
#define helper_psrad_xmm helper_psrad_xmm_x86_64
#define helper_pslld_xmm helper_pslld_xmm_x86_64
#define helper_psrlq_xmm helper_psrlq_xmm_x86_64
#define helper_psllq_xmm helper_psllq_xmm_x86_64
#define helper_psrldq_xmm helper_psrldq_xmm_x86_64
#define helper_pslldq_xmm helper_pslldq_xmm_x86_64
#define helper_paddb_xmm helper_paddb_xmm_x86_64
#define helper_paddw_xmm helper_paddw_xmm_x86_64
#define helper_paddl_xmm helper_paddl_xmm_x86_64
#define helper_paddq_xmm helper_paddq_xmm_x86_64
#define helper_psubb_xmm helper_psubb_xmm_x86_64
#define helper_psubw_xmm helper_psubw_xmm_x86_64
#define helper_psubl_xmm helper_psubl_xmm_x86_64
#define helper_psubq_xmm helper_psubq_xmm_x86_64
#define helper_paddusb_xmm helper_paddusb_xmm_x86_64
#define helper_paddsb_xmm helper_paddsb_xmm_x86_64
#define helper_psubusb_xmm helper_psubusb_xmm_x86_64
#define helper_psubsb_xmm helper_psubsb_xmm_x86_64
#define helper_paddusw_xmm helper_paddusw_xmm_x86_64
#define helper_paddsw_xmm helper_paddsw_xmm_x86_64
#define helper_psubusw_xmm helper_psubusw_xmm_x86_64
#define helper_psubsw_xmm helper_psubsw_xmm_x86_64
#define helper_pminub_xmm helper_pminub_xmm_x86_64
#define helper_pmaxub_xmm helper_pmaxub_xmm_x86_64
#define helper_pminsw_xmm helper_pminsw_xmm_x86_64
#define helper_pmaxsw_xmm helper_pmaxsw_xmm_x86_64
#define helper_pand_xmm helper_pand_xmm_x86_64
#define helper_pandn_xmm helper_pandn_xmm_x86_64
#define helper_por_xmm helper_por_xmm_x86_64
#define helper_pxor_xmm helper_pxor_xmm_x86_64
#define helper_pcmpgtb_xmm helper_pcmpgtb_xmm_x86_64
#define helper_pcmpgtw_xmm helper_pcmpgtw_xmm_x86_64
#define helper_pcmpgtl_xmm helper_pcmpgtl_xmm_x86_64
#define helper_pcmpeqb_xmm helper_pcmpeqb_xmm_x86_64
#define helper_pcmpeqw_xmm helper_pcmpeqw_xmm_x86_64
#define helper_pcmpeql_xmm helper_pcmpeql_xmm_x86_64
#define helper_pmullw_xmm helper_pmullw_xmm_x86_64
#define helper_pmulhuw_xmm helper_pmulhuw_xmm_x86_64
#define helper_pmulhw_xmm helper_pmulhw_xmm_x86_64
#define helper_pavgb_xmm helper_pavgb_xmm_x86_64
#define helper_pavgw_xmm helper_pavgw_xmm_x86_64
#define helper_pmuludq_xmm helper_pmuludq_xmm_x86_64
#define helper_pmaddwd_xmm helper_pmaddwd_xmm_x86_64
#define helper_psadbw_xmm helper_psadbw_xmm_x86_64
#define helper_maskmov_xmm helper_maskmov_xmm_x86_64
#define helper_movl_mm_T0_xmm helper_movl_mm_T0_xmm_x86_64
#define helper_movq_mm_T0_xmm helper_movq_mm_T0_xmm_x86_64
#define helper_shufps helper_shufps_x86_64
#define helper_shufpd helper_shufpd_x86_64
#define helper_pshufd_xmm helper_pshufd_xmm_x86_64
#define helper_pshuflw_xmm helper_pshuflw_xmm_x86_64
#define helper_pshufhw_xmm helper_pshufhw_xmm_x86_64
#define helper_addps helper_addps_x86_64
#define helper_addss helper_addss_x86_64
#define helper_addpd helper_addpd_x86_64
#define helper_addsd helper_addsd_x86_64
#define helper_subps helper_subps_x86_64
#define helper_subss helper_subss_x86_64
#define helper_subpd helper_subpd_x86_64
#define helper_subsd helper_subsd_x86_64
#define helper_mulps helper_mulps_x86_64
#define helper_mulss helper_mulss_x86_64
#define helper_mulpd helper_mulpd_x86_64
#define helper_mulsd helper_mulsd_x86_64
#define helper_divps helper_divps_x86_64
#define helper_divss helper_divss_x86_64
#define helper_divpd helper_divpd_x86_64
#define helper_divsd helper_divsd_x86_64
#define helper_minps helper_minps_x86_64
#define helper_minss helper_minss_x86_64
#define helper_minpd helper_minpd_x86_64
#define helper_minsd helper_minsd_x86_64
#define helper_maxps helper_maxps_x86_64
#define helper_maxss helper_maxss_x86_64
#define helper_maxpd helper_maxpd_x86_64
#define helper_maxsd helper_maxsd_x86_64
#define helper_sqrtps helper_sqrtps_x86_64
#define helper_sqrtss helper_sqrtss_x86_64
#define helper_sqrtpd helper_sqrtpd_x86_64
#define helper_sqrtsd helper_sqrtsd_x86_64
#define helper_cvtps2pd helper_cvtps2pd_x86_64
#define helper_cvtpd2ps helper_cvtpd2ps_x86_64
#define helper_cvtss2sd helper_cvtss2sd_x86_64
#define helper_cvtsd2ss helper_cvtsd2ss_x86_64
#define helper_cvtdq2ps helper_cvtdq2ps_x86_64
#define helper_cvtdq2pd helper_cvtdq2pd_x86_64
#define helper_cvtpi2ps helper_cvtpi2ps_x86_64
#define helper_cvtpi2pd helper_cvtpi2pd_x86_64
#define helper_cvtsi2ss helper_cvtsi2ss_x86_64
#define helper_cvtsi2sd helper_cvtsi2sd_x86_64
#define helper_cvtsq2ss helper_cvtsq2ss_x86_64
#define helper_cvtsq2sd helper_cvtsq2sd_x86_64
#define helper_cvtps2dq helper_cvtps2dq_x86_64
#define helper_cvtpd2dq helper_cvtpd2dq_x86_64
#define helper_cvtps2pi helper_cvtps2pi_x86_64
#define helper_cvtpd2pi helper_cvtpd2pi_x86_64
#define helper_cvtss2si helper_cvtss2si_x86_64
#define helper_cvtsd2si helper_cvtsd2si_x86_64
#define helper_cvtss2sq helper_cvtss2sq_x86_64
#define helper_cvtsd2sq helper_cvtsd2sq_x86_64
#define helper_cvttps2dq helper_cvttps2dq_x86_64
#define helper_cvttpd2dq helper_cvttpd2dq_x86_64
#define helper_cvttps2pi helper_cvttps2pi_x86_64
#define helper_cvttpd2pi helper_cvttpd2pi_x86_64
#define helper_cvttss2si helper_cvttss2si_x86_64
#define helper_cvttsd2si helper_cvttsd2si_x86_64
#define helper_cvttss2sq helper_cvttss2sq_x86_64
#define helper_cvttsd2sq helper_cvttsd2sq_x86_64
#define helper_rsqrtps helper_rsqrtps_x86_64
#define helper_rsqrtss helper_rsqrtss_x86_64
#define helper_rcpps helper_rcpps_x86_64
#define helper_rcpss helper_rcpss_x86_64
#define helper_extrq_r helper_extrq_r_x86_64
#define helper_extrq_i helper_extrq_i_x86_64
#define helper_insertq_r helper_insertq_r_x86_64
#define helper_insertq_i helper_insertq_i_x86_64
#define helper_haddps helper_haddps_x86_64
#define helper_haddpd helper_haddpd_x86_64
#define helper_hsubps helper_hsubps_x86_64
#define helper_hsubpd helper_hsubpd_x86_64
#define helper_addsubps helper_addsubps_x86_64
#define helper_addsubpd helper_addsubpd_x86_64
#define helper_cmpeqps helper_cmpeqps_x86_64
#define helper_cmpeqss helper_cmpeqss_x86_64
#define helper_cmpeqpd helper_cmpeqpd_x86_64
#define helper_cmpeqsd helper_cmpeqsd_x86_64
#define helper_cmpltps helper_cmpltps_x86_64
#define helper_cmpltss helper_cmpltss_x86_64
#define helper_cmpltpd helper_cmpltpd_x86_64
#define helper_cmpltsd helper_cmpltsd_x86_64
#define helper_cmpleps helper_cmpleps_x86_64
#define helper_cmpless helper_cmpless_x86_64
#define helper_cmplepd helper_cmplepd_x86_64
#define helper_cmplesd helper_cmplesd_x86_64
#define helper_cmpunordps helper_cmpunordps_x86_64
#define helper_cmpunordss helper_cmpunordss_x86_64
#define helper_cmpunordpd helper_cmpunordpd_x86_64
#define helper_cmpunordsd helper_cmpunordsd_x86_64
#define helper_cmpneqps helper_cmpneqps_x86_64
#define helper_cmpneqss helper_cmpneqss_x86_64
#define helper_cmpneqpd helper_cmpneqpd_x86_64
#define helper_cmpneqsd helper_cmpneqsd_x86_64
#define helper_cmpnltps helper_cmpnltps_x86_64
#define helper_cmpnltss helper_cmpnltss_x86_64
#define helper_cmpnltpd helper_cmpnltpd_x86_64
#define helper_cmpnltsd helper_cmpnltsd_x86_64
#define helper_cmpnleps helper_cmpnleps_x86_64
#define helper_cmpnless helper_cmpnless_x86_64
#define helper_cmpnlepd helper_cmpnlepd_x86_64
#define helper_cmpnlesd helper_cmpnlesd_x86_64
#define helper_cmpordps helper_cmpordps_x86_64
#define helper_cmpordss helper_cmpordss_x86_64
#define helper_cmpordpd helper_cmpordpd_x86_64
#define helper_cmpordsd helper_cmpordsd_x86_64
#define helper_ucomiss helper_ucomiss_x86_64
#define helper_comiss helper_comiss_x86_64
#define helper_ucomisd helper_ucomisd_x86_64
#define helper_comisd helper_comisd_x86_64
#define helper_movmskps helper_movmskps_x86_64
#define helper_movmskpd helper_movmskpd_x86_64
#define helper_pmovmskb_xmm helper_pmovmskb_xmm_x86_64
#define helper_packsswb_xmm helper_packsswb_xmm_x86_64
#define helper_packuswb_xmm helper_packuswb_xmm_x86_64
#define helper_packssdw_xmm helper_packssdw_xmm_x86_64
#define helper_punpcklbw_xmm helper_punpcklbw_xmm_x86_64
#define helper_punpcklwd_xmm helper_punpcklwd_xmm_x86_64
#define helper_punpckldq_xmm helper_punpckldq_xmm_x86_64
#define helper_punpcklqdq_xmm helper_punpcklqdq_xmm_x86_64
#define helper_punpckhbw_xmm helper_punpckhbw_xmm_x86_64
#define helper_punpckhwd_xmm helper_punpckhwd_xmm_x86_64
#define helper_punpckhdq_xmm helper_punpckhdq_xmm_x86_64
#define helper_punpckhqdq_xmm helper_punpckhqdq_xmm_x86_64
#define helper_pshufb_xmm helper_pshufb_xmm_x86_64
#define helper_phaddw_xmm helper_phaddw_xmm_x86_64
#define helper_phaddd_xmm helper_phaddd_xmm_x86_64
#define helper_phaddsw_xmm helper_phaddsw_xmm_x86_64
#define helper_pmaddubsw_xmm helper_pmaddubsw_xmm_x86_64
#define helper_phsubw_xmm helper_phsubw_xmm_x86_64
#define helper_phsubd_xmm helper_phsubd_xmm_x86_64
#define helper_phsubsw_xmm helper_phsubsw_xmm_x86_64
#define helper_pabsb_xmm helper_pabsb_xmm_x86_64
#define helper_pabsw_xmm helper_pabsw_xmm_x86_64
#define helper_pabsd_xmm helper_pabsd_xmm_x86_64
#define helper_pmulhrsw_xmm helper_pmulhrsw_xmm_x86_64
#define helper_psignb_xmm helper_psignb_xmm_x86_64
#define helper_psignw_xmm helper_psignw_xmm_x86_64
#define helper_psignd_xmm helper_psignd_xmm_x86_64
#define helper_palignr_xmm helper_palignr_xmm_x86_64
#define helper_pblendvb_xmm helper_pblendvb_xmm_x86_64
#define helper_blendvps_xmm helper_blendvps_xmm_x86_64
#define helper_blendvpd_xmm helper_blendvpd_xmm_x86_64
#define helper_ptest_xmm helper_ptest_xmm_x86_64
#define helper_pmovsxbw_xmm helper_pmovsxbw_xmm_x86_64
#define helper_pmovsxbd_xmm helper_pmovsxbd_xmm_x86_64
#define helper_pmovsxbq_xmm helper_pmovsxbq_xmm_x86_64
#define helper_pmovsxwd_xmm helper_pmovsxwd_xmm_x86_64
#define helper_pmovsxwq_xmm helper_pmovsxwq_xmm_x86_64
#define helper_pmovsxdq_xmm helper_pmovsxdq_xmm_x86_64
#define helper_pmovzxbw_xmm helper_pmovzxbw_xmm_x86_64
#define helper_pmovzxbd_xmm helper_pmovzxbd_xmm_x86_64
#define helper_pmovzxbq_xmm helper_pmovzxbq_xmm_x86_64
#define helper_pmovzxwd_xmm helper_pmovzxwd_xmm_x86_64
#define helper_pmovzxwq_xmm helper_pmovzxwq_xmm_x86_64
#define helper_pmovzxdq_xmm helper_pmovzxdq_xmm_x86_64
#define helper_pmuldq_xmm helper_pmuldq_xmm_x86_64
#define helper_pcmpeqq_xmm helper_pcmpeqq_xmm_x86_64
#define helper_packusdw_xmm helper_packusdw_xmm_x86_64
#define helper_pminsb_xmm helper_pminsb_xmm_x86_64
#define helper_pminsd_xmm helper_pminsd_xmm_x86_64
#define helper_pminuw_xmm helper_pminuw_xmm_x86_64
#define helper_pminud_xmm helper_pminud_xmm_x86_64
#define helper_pmaxsb_xmm helper_pmaxsb_xmm_x86_64
#define helper_pmaxsd_xmm helper_pmaxsd_xmm_x86_64
#define helper_pmaxuw_xmm helper_pmaxuw_xmm_x86_64
#define helper_pmaxud_xmm helper_pmaxud_xmm_x86_64
#define helper_pmulld_xmm helper_pmulld_xmm_x86_64
#define helper_phminposuw_xmm helper_phminposuw_xmm_x86_64
#define helper_roundps_xmm helper_roundps_xmm_x86_64
#define helper_roundpd_xmm helper_roundpd_xmm_x86_64
#define helper_roundss_xmm helper_roundss_xmm_x86_64
#define helper_roundsd_xmm helper_roundsd_xmm_x86_64
#define helper_blendps_xmm helper_blendps_xmm_x86_64
#define helper_blendpd_xmm helper_blendpd_xmm_x86_64
#define helper_pblendw_xmm helper_pblendw_xmm_x86_64
#define helper_dpps_xmm helper_dpps_xmm_x86_64
#define helper_dppd_xmm helper_dppd_xmm_x86_64
#define helper_mpsadbw_xmm helper_mpsadbw_xmm_x86_64
#define helper_pcmpgtq_xmm helper_pcmpgtq_xmm_x86_64
#define helper_pcmpestri_xmm helper_pcmpestri_xmm_x86_64
#define helper_pcmpestrm_xmm helper_pcmpestrm_xmm_x86_64
#define helper_pcmpistri_xmm helper_pcmpistri_xmm_x86_64
#define helper_pcmpistrm_xmm helper_pcmpistrm_xmm_x86_64
#define helper_crc32 helper_crc32_x86_64
#define helper_pclmulqdq_xmm helper_pclmulqdq_xmm_x86_64
#define helper_aesdec_xmm helper_aesdec_xmm_x86_64
#define helper_aesdeclast_xmm helper_aesdeclast_xmm_x86_64
#define helper_aesenc_xmm helper_aesenc_xmm_x86_64
#define helper_aesenclast_xmm helper_aesenclast_xmm_x86_64
#define helper_aesimc_xmm helper_aesimc_xmm_x86_64
#define helper_aeskeygenassist_xmm helper_aeskeygenassist_xmm_x86_64
#define cpu_sync_bndcs_hflags cpu_sync_bndcs_hflags_x86_64
#define cpu_x86_support_mca_broadcast cpu_x86_support_mca_broadcast_x86_64
#define x86_cpu_set_a20 x86_cpu_set_a20_x86_64
#define cpu_x86_update_cr0 cpu_x86_update_cr0_x86_64
#define cpu_x86_update_cr3 cpu_x86_update_cr3_x86_64
#define cpu_x86_update_cr4 cpu_x86_update_cr4_x86_64
#define x86_cpu_get_phys_page_attrs_debug x86_cpu_get_phys_page_attrs_debug_x86_64
#define cpu_x86_get_descr_debug cpu_x86_get_descr_debug_x86_64
#define do_cpu_init do_cpu_init_x86_64
#define do_cpu_sipi do_cpu_sipi_x86_64
#define x86_cpu_exec_enter x86_cpu_exec_enter_x86_64
#define x86_cpu_exec_exit x86_cpu_exec_exit_x86_64
#define x86_ldub_phys x86_ldub_phys_x86_64
#define x86_lduw_phys x86_lduw_phys_x86_64
#define x86_ldl_phys x86_ldl_phys_x86_64
#define x86_ldq_phys x86_ldq_phys_x86_64
#define x86_stb_phys x86_stb_phys_x86_64
#define x86_stl_phys_notdirty x86_stl_phys_notdirty_x86_64
#define x86_stw_phys x86_stw_phys_x86_64
#define x86_stl_phys x86_stl_phys_x86_64
#define x86_stq_phys x86_stq_phys_x86_64
#define helper_divb_AL helper_divb_AL_x86_64
#define helper_idivb_AL helper_idivb_AL_x86_64
#define helper_divw_AX helper_divw_AX_x86_64
#define helper_idivw_AX helper_idivw_AX_x86_64
#define helper_divl_EAX helper_divl_EAX_x86_64
#define helper_idivl_EAX helper_idivl_EAX_x86_64
#define helper_aam helper_aam_x86_64
#define helper_aad helper_aad_x86_64
#define helper_aaa helper_aaa_x86_64
#define helper_aas helper_aas_x86_64
#define helper_daa helper_daa_x86_64
#define helper_das helper_das_x86_64
#define helper_divq_EAX helper_divq_EAX_x86_64
#define helper_idivq_EAX helper_idivq_EAX_x86_64
#define helper_pdep helper_pdep_x86_64
#define helper_pext helper_pext_x86_64
#define helper_rclb helper_rclb_x86_64
#define helper_rcrb helper_rcrb_x86_64
#define helper_rclw helper_rclw_x86_64
#define helper_rcrw helper_rcrw_x86_64
#define helper_rcll helper_rcll_x86_64
#define helper_rcrl helper_rcrl_x86_64
#define helper_rclq helper_rclq_x86_64
#define helper_rcrq helper_rcrq_x86_64
#define helper_cr4_testbit helper_cr4_testbit_x86_64
#define helper_rdrand helper_rdrand_x86_64
#define helper_cmpxchg8b_unlocked helper_cmpxchg8b_unlocked_x86_64
#define helper_cmpxchg8b helper_cmpxchg8b_x86_64
#define helper_cmpxchg16b_unlocked helper_cmpxchg16b_unlocked_x86_64
#define helper_cmpxchg16b helper_cmpxchg16b_x86_64
#define helper_boundw helper_boundw_x86_64
#define helper_boundl helper_boundl_x86_64
#define helper_outb helper_outb_x86_64
#define helper_inb helper_inb_x86_64
#define helper_outw helper_outw_x86_64
#define helper_inw helper_inw_x86_64
#define helper_outl helper_outl_x86_64
#define helper_inl helper_inl_x86_64
#define helper_into helper_into_x86_64
#define helper_cpuid helper_cpuid_x86_64
#define helper_read_crN helper_read_crN_x86_64
#define helper_write_crN helper_write_crN_x86_64
#define helper_lmsw helper_lmsw_x86_64
#define helper_invlpg helper_invlpg_x86_64
#define helper_rdtsc helper_rdtsc_x86_64
#define helper_rdtscp helper_rdtscp_x86_64
#define helper_rdpmc helper_rdpmc_x86_64
#define helper_wrmsr helper_wrmsr_x86_64
#define helper_rdmsr helper_rdmsr_x86_64
#define helper_hlt helper_hlt_x86_64
#define helper_monitor helper_monitor_x86_64
#define helper_mwait helper_mwait_x86_64
#define helper_pause helper_pause_x86_64
#define helper_debug helper_debug_x86_64
#define helper_rdpkru helper_rdpkru_x86_64
#define helper_wrpkru helper_wrpkru_x86_64
#define helper_bndck helper_bndck_x86_64
#define helper_bndldx64 helper_bndldx64_x86_64
#define helper_bndldx32 helper_bndldx32_x86_64
#define helper_bndstx64 helper_bndstx64_x86_64
#define helper_bndstx32 helper_bndstx32_x86_64
#define helper_bnd_jmp helper_bnd_jmp_x86_64
#define helper_syscall helper_syscall_x86_64
#define helper_sysret helper_sysret_x86_64
#define x86_cpu_do_interrupt x86_cpu_do_interrupt_x86_64
#define do_interrupt_x86_hardirq do_interrupt_x86_hardirq_x86_64
#define x86_cpu_exec_interrupt x86_cpu_exec_interrupt_x86_64
#define helper_lldt helper_lldt_x86_64
#define helper_ltr helper_ltr_x86_64
#define uc_check_cpu_x86_load_seg uc_check_cpu_x86_load_seg_x86_64
#define helper_load_seg helper_load_seg_x86_64
#define helper_ljmp_protected helper_ljmp_protected_x86_64
#define helper_lcall_real helper_lcall_real_x86_64
#define helper_lcall_protected helper_lcall_protected_x86_64
#define helper_iret_real helper_iret_real_x86_64
#define helper_iret_protected helper_iret_protected_x86_64
#define helper_lret_protected helper_lret_protected_x86_64
#define helper_sysenter helper_sysenter_x86_64
#define helper_sysexit helper_sysexit_x86_64
#define helper_lsl helper_lsl_x86_64
#define helper_lar helper_lar_x86_64
#define helper_verr helper_verr_x86_64
#define helper_verw helper_verw_x86_64
#define cpu_x86_load_seg cpu_x86_load_seg_x86_64
#define helper_check_iob helper_check_iob_x86_64
#define helper_check_iow helper_check_iow_x86_64
#define helper_check_iol helper_check_iol_x86_64
#define do_smm_enter do_smm_enter_x86_64
#define helper_rsm helper_rsm_x86_64
#define helper_vmrun helper_vmrun_x86_64
#define helper_vmmcall helper_vmmcall_x86_64
#define helper_vmload helper_vmload_x86_64
#define helper_vmsave helper_vmsave_x86_64
#define helper_stgi helper_stgi_x86_64
#define helper_clgi helper_clgi_x86_64
#define helper_skinit helper_skinit_x86_64
#define helper_invlpga helper_invlpga_x86_64
#define cpu_svm_check_intercept_param cpu_svm_check_intercept_param_x86_64
#define helper_svm_check_intercept_param helper_svm_check_intercept_param_x86_64
#define helper_svm_check_io helper_svm_check_io_x86_64
#define cpu_vmexit cpu_vmexit_x86_64
#define do_vmexit do_vmexit_x86_64
#define tcg_x86_init tcg_x86_init_x86_64
#define gen_intermediate_code gen_intermediate_code_x86_64
#define restore_state_to_opc restore_state_to_opc_x86_64
#define x86_cpu_xsave_all_areas x86_cpu_xsave_all_areas_x86_64
#define x86_cpu_xrstor_all_areas x86_cpu_xrstor_all_areas_x86_64
#define cpu_get_fp80 cpu_get_fp80_x86_64
#define cpu_set_fp80 cpu_set_fp80_x86_64
#endif
