/* Automatically generated by create_config - do not modify */
#define TARGET_S390X 1
#define TARGET_NAME "s390x"
#define TARGET_S390X 1
#define TARGET_SYSTBL_ABI common,64
#define TARGET_WORDS_BIGENDIAN 1
#define CONFIG_SOFTMMU 1
#define TARGET_SUPPORTS_MTTCG 1
